"""
    Optimizer.JobStateCanvas.py

    Copyright (c) 2021-2024, SAXS Team, KEK-PF
"""
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.gridspec import GridSpec
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from datetime import timedelta
from molass_legacy.KekLib.OurTkinter import Tk
from molass_legacy.KekLib.OurMatplotlib import NavigationToolbar
from molass_legacy.KekLib.TkUtils import split_geometry
from molass_legacy._MOLASS.SerialSettings import get_setting
from molass_legacy.SerialAnalyzer.DataUtils import get_in_folder
from molass_legacy.KekLib.NumpyUtils import get_proportional_points
from molass_legacy.KekLib.BasicUtils import ordinal_str
from molass_legacy.KekLib.TimeUtils import friendly_time_str
from molass_legacy.KekLib.ExceptionTracebacker import log_exception
from .NaviFrame import NaviFrame
from .FvScoreConverter import convert_score

SHOW_FV_MIN = 0
SHOW_FV_MAX = 100
PROGRESS_X_MARGIN = 1000    # not too large for xmax = 500000
DEVELOP_MODE = True
ENABLE_DEVEL_POPUP_WHILE_BUSY = True

class JobStateCanvas(Tk.Frame):
    def __init__(self, parent, dialog):
        self.parent = parent
        self.dialog = dialog
        self.elution_model = get_setting("elution_model")
        self.update_init_state(dialog.state_info)
        self.iconified = False
        self.logger = dialog.logger
        self.optinit_info = dialog.optinit_info
        self.composite = self.optinit_info.composite
        self.suptitle = None
        Tk.Frame.__init__(self, parent)

        cframe = Tk.Frame(self)
        cframe.pack()
        tframe = Tk.Frame(self)
        tframe.pack(fill=Tk.X, padx=10)
        tframe_left = Tk.Frame(tframe)
        tframe_left.pack(side=Tk.LEFT)
        navi_frame = NaviFrame(tframe, self)
        navi_frame.pack(side=Tk.RIGHT)

        self.fig = fig = plt.figure(figsize=(18, 9))
        gs = GridSpec(33, 15)
        axes = []
        for j in range(3):
            j_ = j*5
            ax = fig.add_subplot(gs[0:16,j_:j_+5])
            axes.append(ax)

        axt = axes[1].twinx()
        axes.append(axt)
        self.axes = axes
        self.prog_ax = fig.add_subplot(gs[17:21,2:])
        peak_ax = fig.add_subplot(gs[21:25,2:])
        rg_ax = fig.add_subplot(gs[25:29,2:])
        map_ax = fig.add_subplot(gs[29:33,2:])
        self.prog_axes = [self.prog_ax, peak_ax, rg_ax, map_ax]
        for ax in self.prog_axes[0:3]:
            ax.set_xticklabels([])
        self.prog_title_axes = [fig.add_subplot(gs[17+i*4:21+i*4,0:2]) for i in range(0,4)]
        prog_titles = ["Function SV", "Peak Top Positions", "Rg Values", "Mapped Range"]
        for ax, title in zip(self.prog_title_axes, prog_titles):
            ax.set_axis_off()
            ax.text(-0.3, 0.5, title, fontsize=16)
        self.mpl_canvas = FigureCanvasTkAgg(self.fig, cframe)
        self.mpl_canvas_widget = self.mpl_canvas.get_tk_widget()
        self.mpl_canvas_widget.pack(fill=Tk.BOTH, expand=1)
        self.optview_ranges = None
        self.initial_view_ranges = None

        minimize_btn = Tk.Button(self.mpl_canvas_widget, bg="white", text="Minimize Window", command=self.minimize_window)
        minimize_btn.place(relx=0.925, rely=0)   # anchor=Tk.NE?

        button_text = self.get_view_width_button_text()
        self.view_width_btn = Tk.Button(self.mpl_canvas_widget, bg="white", text=button_text, command=self.change_view_widths)
        self.view_width_btn.place(relx=0.94, rely=0.50)

        residual_view_btn = Tk.Button(self.mpl_canvas_widget, bg="white", text="Residual View", command=self.show_residual_view)
        residual_view_btn.place(relx=0.94, rely=0.55)

        boundary_view_btn = Tk.Button(self.mpl_canvas_widget, bg="white", text="Boundary View", command=self.show_boundary_view)
        boundary_view_btn.place(relx=0.935, rely=0.60)

        if False:
            state = Tk.NORMAL if self.elution_model == 0 else Tk.DISABLED
            sec_inspect_btn = Tk.Button(self.mpl_canvas_widget, bg="white", text="SEC Conformance", command=self.show_sec_conformance, state=state)
            sec_inspect_btn.place(relx=0.925, rely=0.65)

        self.toolbar = NavigationToolbar(self.mpl_canvas, tframe)
        self.toolbar.update()
        self.popup_menu_0 = None
        self.popup_menu_1 = None
        self.popup_menu_prog = None
        self.popup_menu_devel = None
        self.mpl_canvas.mpl_connect('button_press_event', self.on_figure_click)

    def get_view_width_button_text(self):
        if self.optview_ranges is None:
            text = "Narrower View"
        else:
            text = "Wider View"
        return text

    def update_view_width_button(self):
        self.view_width_btn.config(text=self.get_view_width_button_text())

    def update_init_state(self, state_info):
        self.state_info = state_info
        self.dsets = state_info.dsets
        self.demo_index = state_info.demo_index
        self.demo_info = state_info.demo_info
        self.fullopt = state_info.fullopt

    def close(self):
        plt.close(self.fig)
        self.on_focus_in()  # to focus to the parent dialog

    def minimize_window(self):
        self.dialog.iconify()
        self.dialog.parent_dialog.iconify()
        self.iconified =True
        self.bind("<FocusIn>", self.on_focus_in)

    def on_focus_in(self, *args):
        if self.iconified:
            print("deiconify")
            self.dialog.parent_dialog.deiconify()
            self.iconified =False

    def draw_main(self):
        if self.demo_index is None:
            self.demo_info = None
            self.draw_state()
        else:
            self.draw_state()
            self.draw_progress()
        self.fig.subplots_adjust(top=0.9, left=0.05, right=0.9, hspace=1.1, wspace=1.5, bottom=0.05)

    def draw_state(self):
        if self.demo_index is None:
            self.draw_suptitle()
        else:
            """
            refactor with set_demo_info, draw_state_impl, etc.
            """
            i = self.get_best_index()
            self.best_index = i
            self.curr_index = i
            self.draw_indexed_state(i)

    def draw_suptitle(self):
        from molass_legacy.Optimizer.OptimizerUtils import get_model_name, get_method_name
        job_info = self.dialog.get_job_info()
        job_name = job_info[0]
        in_folder = get_in_folder()
        model_name = get_model_name(self.dialog.class_code)
        text = "Job %s State at %s local minimum on %s with model=%s method=%s" % (
            job_name, ordinal_str(self.curr_index), in_folder, model_name, get_method_name())
        if self.suptitle is None:
            self.suptitle = self.fig.suptitle(text, fontsize=20)
        else:
            self.suptitle.set_text(text)

    def draw_progress(self):
        for ax in self.prog_axes:
            ax.cla()
        for ax in self.prog_axes[0:3]:
            ax.set_xticklabels([])

        # Function Values
        prog_ax = self.prog_ax

        fv, max_num_evals = self.get_fv_array()
        x_, y_ = fv[:,0:2].T
        prog_ax.plot(x_, convert_score(y_))
        prog_ax.set_xlim(-PROGRESS_X_MARGIN, max_num_evals + PROGRESS_X_MARGIN)
        ymin, ymax = prog_ax.get_ylim()
        prog_ax.set_ylim(ymin, ymax)     # these limits will be reset below

        m = self.best_index
        ymin, ymax = prog_ax.get_ylim()
        prog_ax.set_ylim(max(SHOW_FV_MIN, ymin), SHOW_FV_MAX)

        params_type = self.fullopt.params_type

        # Peak Top Positions
        peak_ax = self.prog_axes[1]
        x_array = self.demo_info[1]
        n = self.fullopt.n_components
        x_ = fv[:,0]
        pos_array_list = params_type.get_peak_pos_array_list(x_array)
        for y_ in pos_array_list:
            peak_ax.plot(x_, y_)

        xmin, xmax = self.axes[1].get_xlim()
        xmin_, xmax_ = get_proportional_points(xmin, xmax, [-0.1, 1.1])
        peak_ax.set_ylim(xmin_, xmax_)

        # Rg Values
        rg_ax = self.prog_axes[2]
        rg_start = params_type.get_rg_start_index()
        for k in range(n):
            y_ = x_array[:,rg_start+k]
            rg_ax.plot(x_, y_)
        ymin_rg, ymax_rg = rg_ax.get_ylim()
        rg_ax.set_ylim(10, ymax_rg*1.2)

        # Mapped Range
        map_ax = self.prog_axes[3]
        mr_start = params_type.get_mr_start_index()
        for k in range(2):
            y_ = x_array[:,mr_start+k]
            map_ax.plot(x_, y_)
        map_ax.set_ylim(xmin_, xmax_)

        # Best and Current Result Indicator
        xmin, xmax = self.get_xlim_prog_axes()
        best_x = fv[m,0]
        for k, ax in enumerate(self.prog_axes):
            if k > 0:
                ax.set_xlim(xmin, xmax)
                ymin, ymax = ax.get_ylim()
                ax.set_ylim(ymin, ymax)
            ax.plot([best_x, best_x], [ymin, ymax], color='red')
            if self.curr_index != m:
                x = fv[self.curr_index,0]
                ax.plot([x, x], [ymin, ymax], color='yellow')

            if fv.shape[0]-1 not in [m, self.curr_index]:
                x = fv[-1,0]
                ax.plot([x, x], [ymin, ymax], color='gray', alpha=0.3)

        ymin, ymax = map_ax.get_ylim()
        tx = xmax*1.07

        w = 2.2
        ty = ymin*(1-w) + ymax*w
        map_ax.text(tx, ty, "Starting Time", ha="center")

        time_str = self.get_time_started()
        w = 2.0
        ty = ymin*(1-w) + ymax*w
        map_ax.text(tx, ty, time_str, ha="center", va="center")

        w = 1.6
        ty = ymin*(1-w) + ymax*w
        map_ax.text(tx, ty, "Time Elapsed", ha="center")

        time_str = self.get_time_elapsed()
        w = 1.4
        ty = ymin*(1-w) + ymax*w
        map_ax.text(tx, ty, time_str, ha="center", va="center")

        w = 0.4
        ty = ymin*(1-w) + ymax*w
        map_ax.text(tx, ty, "Ending Time", ha="center")

        # guess_ending_time() must be called before get_remaining_time()
        time_str = self.guess_ending_time(fv)
        w = 0.2
        ty = ymin*(1-w) + ymax*w
        map_ax.text(tx, ty, time_str, ha="center", va="center")

        w = 1.0
        ty = ymin*(1-w) + ymax*w
        map_ax.text(tx, ty, "Time Ahead", ha="center")

        time_str = self.get_remaining_time()
        w = 0.8
        ty = ymin*(1-w) + ymax*w
        map_ax.text(tx, ty, time_str, ha="center", va="center")

    def get_xlim_prog_axes(self):
        fv, max_num_evals = self.get_fv_array()
        xmin = -PROGRESS_X_MARGIN
        xmax = max_num_evals + PROGRESS_X_MARGIN
        return xmin, xmax

    def set_demo_info(self, demo_info):
        self.demo_info = demo_info
        self.curr_fv = self.demo_info[0]
        self.draw_suptitle()
        i = self.get_best_index()
        self.best_index = i
        self.curr_index = i
        self.draw_state_impl(i)
        self.draw_progress()

    def get_running_solver_info(self):
        return self.dialog.runner.solver, self.dialog.optinit_info.n_iterations

    def get_fv_array(self):
        if self.demo_info is None:
            assert False
            xmax = 500000
        else:
            fv = self.demo_info[0]
            solver_name, niter = self.get_running_solver_info()
            if solver_name == "ultranest":
                from molass_legacy.Solvers.UltraNest.SolverUltraNest import get_max_ncalls
                # task: unify this estimation
                xmax = get_max_ncalls(niter)
            else:
                xmax = self.demo_info[2]
        return fv, xmax

    def draw_indexed_state(self, i):
        self.draw_suptitle()

        fv = convert_score(self.demo_info[0][i][1])

        for ax in self.axes:
            ax.cla()
        ax1, ax2, ax3, axt = self.axes
        axt.grid(False)
        ax1.set_title("UV Decomposition", fontsize=16)
        ax2.set_title("Xray Decomposition", fontsize=16)
        ax3.set_title("Objective Function Scores in SV=%.3g" % fv, fontsize=16)
        x_array = self.demo_info[1]
        p = x_array[i,:] 
        self.fullopt.objective_func(p, plot=True, axis_info=(self.fig, self.axes))
        self.set_view_widths()

    def get_best_index(self, stop=None):
        fv, xmax = self.get_fv_array()
        m = np.argmin(fv[:stop,1])
        return m

    def get_curr_index(self):
        return self.curr_index

    def get_first(self):
        self.curr_index = 0
        self.draw_state_impl(self.curr_index)

    def get_last(self):
        fv = self.demo_info[0]
        self.curr_index = fv.shape[0] - 1
        self.draw_state_impl(self.curr_index)

    def get_previous(self):
        if self.curr_index > 0:
            self.curr_index -= 1
        self.draw_state_impl(self.curr_index)

    def get_previous_best(self):
        i = self.get_best_index(stop=self.curr_index)
        self.curr_index = i
        self.draw_state_impl(self.curr_index)

    def get_best(self):
        self.curr_index = self.get_best_index()
        self.draw_state_impl(self.curr_index)

    def get_next(self):
        fv, xmax = self.get_fv_array()
        if self.curr_index < fv.shape[0] - 1:
            self.curr_index += 1
        self.draw_state_impl(self.curr_index)

    def get_next_best(self):
        fv, xmax = self.get_fv_array()
        stop = min(len(fv), self.curr_index + 1)
        best_fv_until_now = np.min(fv[:stop,1])
        w  = np.where(fv[stop:,1] < best_fv_until_now)[0]
        if len(w) > 0:
            self.curr_index = stop + w[0]
            self.draw_state_impl(self.curr_index)

    def get_time_started(self):
        try:
            fv_array = self.demo_info[0]
            start_time = fv_array[0,3]
            time = friendly_time_str(start_time)
        except:
            # IndexError: index 3 is out of bounds for axis 1 with size 2
            time = ""
        return time

    def get_time_elapsed(self):
        try:
            fv_array = self.demo_info[0]
            start_time = fv_array[0,3]
            curr_time = fv_array[-1,3]
            hhmmss = str(curr_time - start_time).split(":")
            time = "%3d.%02d" % tuple([int(s) for s in hhmmss[0:2]])
            # %3d instead of %2d is just for positioning purpose with non-fixed-width fonts.
        except:
            log_exception(self.logger, "get_time_elapsed: ")
            # IndexError: index 3 is out of bounds for axis 1 with size 2
            time = ""
        return time

    def guess_ending_time(self, fv):
        finish_time = None
        time = ""

        if fv.shape[0] > 3:
            # unreliable when fv.shape[0] <= 3
            if self.demo_index == 0:
                self.num_iter = self.dialog.num_iter.get()
                try:
                    fv_array = self.demo_info[0]
                    start_time = fv_array[0,3]
                    curr_time = fv_array[-1,3]
                    finish_time = start_time + (curr_time - start_time)*(self.num_iter/fv_array.shape[0])
                    # add 1 minute so that it won't be too early
                    time = friendly_time_str(finish_time + timedelta(minutes=1))
                except:
                    pass

        self.finish_time = finish_time
        return time

    def get_remaining_time(self):
        if self.finish_time is None:
            return ""

        try:
            fv_array = self.demo_info[0]
            curr_time = fv_array[-1,3]
            # add 1 minute so that it won't be too short
            hhmmss = str(self.finish_time - curr_time + timedelta(minutes=1) ).split(":")
            time = "%3d.%02d" % tuple([int(s) for s in hhmmss[0:2]])
            # %3d instead of %2d is just for positioning purpose with non-fixed-width fonts.
            # to be fixed: ValueError: invalid literal for int() with base 10: '-1 day, 23'
        except:
            # log_exception(self.logger, "get_remaining_time: ")
            # IndexError: index 3 is out of bounds for axis 1 with size 2
            # ValueError: invalid literal for int() with base 10: '-1 day, 23'
            time = ""
        return time

    def draw_state_impl(self, i):
        print("draw_state_impl: i=", i)
        self.draw_indexed_state(i)
        self.draw_progress()
        self.mpl_canvas.draw()

    def get_params(self, index):
        return self.demo_info[1][index]

    def get_current_params(self):
        return self.get_params(self.curr_index)

    def get_best_params(self):
        best_index = self.get_best_index()
        return self.get_params(best_index)

    def get_current_fv(self):
        fv, xmax = self.get_fv_array()
        return fv[self.curr_index][1]

    def show_params(self, debug=True):
        if debug:
            from importlib import reload
            import Optimizer.ParamsInspection
            reload(Optimizer.ParamsInspection)
        from .ParamsInspection import ParamsInspection
        self.dialog.grab_set()  # temporary fix to the grab_release problem
        params = self.get_current_params()
        dialog = ParamsInspection(self.dialog.parent, params, self.dsets, self.fullopt, state_info=self.state_info)
        dialog.show()
        self.dialog.grab_set()  # temporary fix to the grab_release problem

    def show_complementary_view(self, debug=True):
        if debug or DEVELOP_MODE:
            from importlib import reload
            import Optimizer.ComplementaryView
            reload(Optimizer.ComplementaryView)
        from molass_legacy.Optimizer.ComplementaryView import ComplementaryView

        self.dialog.grab_set()  # temporary fix to the grab_release problem

        params = self.demo_info[1][self.curr_index]
        work_folder = self.dialog.get_curr_work_folder()
        cv = ComplementaryView(self.dialog.parent, self.fullopt, self.curr_index, params, work_folder, sd=self.optinit_info.sd)
        cv.show()

        self.dialog.grab_set()  # temporary fix to the grab_release problem

    def show_2p_analysis(self, debug=True):
        if debug:
            from importlib import reload
            import Optimizer.TwoParamAnalysis
            reload(Optimizer.TwoParamAnalysis)
        from .TwoParamAnalysis import TwoParamAnalysis

        self.dialog.grab_set()  # temporary fix to the grab_release problem
        dialog = TwoParamAnalysis(self.dialog.parent, self.dsets, self.dialog.fullopt, self.demo_info, self.curr_index)
        dialog.show()
        self.dialog.grab_set()  # temporary fix to the grab_release problem

    def save_the_figure(self, fig_file):
        self.fig.savefig(fig_file)

    def show_residual_view(self):
        from .ResidualView import ResidualView
        view = ResidualView(self.dialog.parent, self.dialog, self)
        view.show()

    def show_boundary_view(self, debug=True):
        if debug:
            from importlib import reload
            import Optimizer.BoundaryView
            reload(Optimizer.BoundaryView)
        from .BoundaryView import BoundaryView
        view = BoundaryView(self.dialog.parent, self.dialog, self)
        view.show()

    def show_sec_conformance(self, debug=True):
        if debug:
            from importlib import reload
            import Optimizer.SecConformance
            reload(Optimizer.SecConformance)
        from .SecConformance import SecConformance
        inspect = SecConformance(self.dialog.parent, self.dialog, self)
        inspect.show()

    def on_figure_click(self, event):
        if event.button == 3:
            if event.inaxes in [self.axes[0]]:
                self.show_popup_menu_0(event)
            elif event.inaxes in [self.axes[1], self.axes[3]]:
                self.show_popup_menu_1(event)
            else:
                if event.inaxes in self.prog_axes:
                    if ENABLE_DEVEL_POPUP_WHILE_BUSY:
                         self.show_popup_menu_prog(event)
                    else:
                        if not self.dialog.is_busy():
                            self.show_popup_menu_prog(event)
                elif event.inaxes == self.axes[2]:
                    self.show_popup_menu_devel(event)

    def show_popup_menu_0(self, event):
        from molass_legacy.KekLib.PopupMenuUtils import post_popup_menu
        self.create_popup_menu_0(event)
        post_popup_menu(self.popup_menu_0, self.mpl_canvas_widget, event, mpl_event=True)

    def create_popup_menu_0(self, event):
        if self.popup_menu_0 is None:
            self.popup_menu_0 = Tk.Menu(self, tearoff=0 )
            self.popup_menu_0.add_command(label='Show Simple 3D View', command=self.show_simple_3d_view)

    def show_popup_menu_1(self, event):
        self.create_popup_menu_1(event)
        canvas = self.mpl_canvas_widget
        cx = canvas.winfo_rootx()
        cy = canvas.winfo_rooty()
        w, h, x, y = split_geometry(canvas.winfo_geometry())
        self.popup_menu_1.post(cx + int(event.x), cy + h - int(event.y))

    def create_popup_menu_1(self, event):
        from molass_legacy._MOLASS.Version import is_developing_version
        if self.popup_menu_1 is None:
            self.popup_menu_1 = Tk.Menu(self, tearoff=0 )
            self.popup_menu_1.add_command(label='Show This Figure', command=self.show_this_figure)
            self.popup_menu_1.add_command(label='Show Rg-visible Figure', command=self.show_rg_visible_figure)
            self.popup_menu_1.add_command(label='Show Rg-visible Figure with SEC Range', command=lambda: self.show_rg_visible_figure(with_range=True))
            self.popup_menu_1.add_command(label='Show Simple 3D View', command=self.show_simple_3d_view)
            self.popup_menu_1.add_command(label='Show Asc/Dsc-Difference', command=self.show_asc_dsc_difference)
            if is_developing_version():
                self.popup_menu_1.add_separator()
                self.popup_menu_1.add_command(label='Investigate SDM Params', command=self.show_sdm_investigator)
                self.popup_menu_1.add_command(label='Investigate SDM LRF Params', command=self.show_sdm_lrf_investigator)
                self.popup_menu_1.add_command(label='Show Ad hoc Figure', command=self.show_adhoc_figure)


    def show_popup_menu_prog(self, event):    
        self.create_popup_menu_prog(event)
        canvas = self.mpl_canvas_widget
        cx = canvas.winfo_rootx()
        cy = canvas.winfo_rooty()
        w, h, x, y = split_geometry(canvas.winfo_geometry())
        self.popup_menu_prog.post(cx + int(event.x), cy + h - int(event.y))

    def create_popup_menu_prog(self, event):
        if self.popup_menu_prog is None:
            self.popup_menu_prog = Tk.Menu(self, tearoff=0 )
            self.popup_menu_prog.add_command(label='Draw all Parameters', command=self.draw_all_parameters)
            if self.elution_model == 1:
                self.popup_menu_prog.add_command(label='SEC Inspection', command=self.sec_inspection)
            self.popup_menu_prog.add_command(label='SEC Conformance Demo', command=self.sec_conformance_demo)
            self.popup_menu_prog.add_command(label='Function Debugger', command=lambda: self.show_function_debugger(composite=self.fullopt.composite))
            self.popup_menu_prog.add_command(label='Function Debugger (no composite)', command=self.show_function_debugger)
            self.popup_menu_prog.add_command(label='FV Score Inspection', command=self.show_fvscore_inspector)
            # self.popup_menu_prog.add_command(label='Compare Scores', command=self.compare_scores)

    def show_popup_menu_devel(self, event):
        self.create_popup_menu_devel(event)
        canvas = self.mpl_canvas_widget
        cx = canvas.winfo_rootx()
        cy = canvas.winfo_rooty()
        w, h, x, y = split_geometry(canvas.winfo_geometry())
        self.popup_menu_devel.post(cx + int(event.x), cy + h - int(event.y))

    def create_popup_menu_devel(self, event):
        if self.popup_menu_devel is None:
            self.popup_menu_devel = Tk.Menu(self, tearoff=0 )
            self.popup_menu_devel.add_command(label='Show Score Details', command=self.show_score_details)
            self.popup_menu_devel.add_command(label='Inspect Score Transition', command=self.inspect_score_transition)
            self.popup_menu_devel.add_command(label='Update Guinier Region', command=self.test_update_guinier_region)
            self.popup_menu_devel.add_command(label='Show Guinier Deviation', command=self.show_guinier_deviation)
            if not self.dialog.is_busy():
                self.popup_menu_devel.add_command(label='Change Function', command=self.change_function)

    def show_rg_visible_figure(self, with_range=False, debug=True):
        if debug:
            import Optimizer.RgVisibleFigure
            from importlib import reload
            reload(Optimizer.RgVisibleFigure)
        from molass_legacy.Optimizer.RgVisibleFigure import RgVisibleFigure
        dialog = RgVisibleFigure(self.parent, self, with_range=with_range)
        dialog.show()

    def show_simple_3d_view(self, debug=True):
        if debug:
            import Tools.SimpleThreedView
            from importlib import reload
            reload(Tools.SimpleThreedView)
        from molass_legacy.Tools.SimpleThreedView import show_simple_3d_view
        show_simple_3d_view(self.optinit_info.sd)

    def show_asc_dsc_difference(self, debug=True):
        if debug:
            import Optimizer.AscDscDifference
            from importlib import reload
            reload(Optimizer.AscDscDifference)
        from molass_legacy.Optimizer.AscDscDifference import show_asc_dsc_difference
        show_asc_dsc_difference(self)

    def draw_all_parameters(self, debug=False):
        if debug:
            import Optimizer.AllParameters
            from importlib import reload
            reload(Optimizer.AllParameters)
        from molass_legacy.Optimizer.AllParameters import AllParameters
        fv, max_num_evals = self.get_fv_array()
        xmin, xmax = self.get_xlim_prog_axes()
        plot = AllParameters(self.parent, self.fullopt, self.demo_info[1], self.curr_index, self.best_index, fv[:,0], xmin, xmax)
        plot.show()

    def sec_inspection(self, debug=True):
        if debug:
            import Optimizer.SecInspection
            from importlib import reload
            reload(Optimizer.SecInspection)
        from molass_legacy.Optimizer.SecInspection import SecInspection
        dialog = SecInspection(self.parent, self)
        dialog.show()

    def sec_conformance_demo(self, debug=True):
        if debug:
            import SecTheory.ConformanceDemo
            from importlib import reload
            reload(SecTheory.ConformanceDemo)
        from SecTheory.ConformanceDemo import ConformanceDemo
        dialog = ConformanceDemo(self.parent)
        dialog.show()

    def show_function_debugger(self, composite=None, debug=True):
        if debug:
            import Optimizer.FunctionDebugger
            from importlib import reload
            reload(Optimizer.FunctionDebugger)
        from molass_legacy.Optimizer.FunctionDebugger import FunctionDebugger
        debugger = FunctionDebugger(self.parent, self.dialog, self, composite=composite)
        debugger.show()

    def show_fvscore_inspector(self, debug=True):
        if debug:
            import Optimizer.FvScoreInspecor
            from importlib import reload
            reload(Optimizer.FvScoreInspecor)
        from molass_legacy.Optimizer.FvScoreInspecor import FvScoreInspecor
        designer = FvScoreInspecor(self.parent, self)
        designer.show()

    def inspect_score_transition(self, debug=True):
        if debug:
            import Optimizer.ScoreTransition
            from importlib import reload
            reload(Optimizer.ScoreTransition)
        from molass_legacy.Optimizer.ScoreTransition import ScoreTransition
        fv, xmax = self.get_fv_array()
        compare = ScoreTransition(self.dialog, self, self.fullopt, self.demo_info[1], fv, self.best_index)
        compare.show()

    def show_score_details(self, debug=True):
        if debug:
            import Optimizer.FvScoreDetails as details_module
            from importlib import reload
            reload(details_module)
        from .FvScoreDetails import FvScoreDetails
        params = self.get_current_params()
        details = FvScoreDetails(self.parent, self.fullopt, params)
        details.show()

    def test_update_guinier_region(self, debug=True):
        if debug:
            import GuinierTools.GuinierDeviationTester
            from importlib import reload
            reload(GuinierTools.GuinierDeviationTester)
        from ..GuinierTools.GuinierDeviationTester import test_update_guinier_region_impl
        test_update_guinier_region_impl(self)

    def show_guinier_deviation(self, debug=True):
        if debug:
            import GuinierTools.GuinierDeviationTester
            from importlib import reload
            reload(GuinierTools.GuinierDeviationTester)
        from ..GuinierTools.GuinierDeviationTester import show_guinier_deviation
        show_guinier_deviation(self)

    def change_function(self, debug=False):
        if debug:
            import Optimizer.FunctionChanger
            from importlib import reload
            reload(Optimizer.FunctionChanger)
        from molass_legacy.Optimizer.FunctionChanger import FunctionChanger
        dialog = FunctionChanger(self.parent, self)
        dialog.show()

    def show_vp_analysis(self, modelname, debug=True):
        if debug:
            import V2PropOptimizer.VariedPropAnalysis
            from importlib import reload
            reload(V2PropOptimizer.VariedPropAnalysis)
        from V2PropOptimizer.VariedPropAnalysis import VariedPropAnalysis
        dialog = VariedPropAnalysis(self.parent, self, modelname)
        dialog.show()

    def show_mw_integrity(self, debug=True):
        if debug:
            import Optimizer.MwIntegrity
            from importlib import reload
            reload(Optimizer.MwIntegrity)
        from molass_legacy.Optimizer.MwIntegrity import MwIntegrityPlot
        dialog = MwIntegrityPlot(self.parent, self)
        dialog.show()

    def show_sdm_investigator(self, debug=True):
        if debug:
            import Models.Stochastic.DispersiveUtils
            from importlib import reload
            reload(Models.Stochastic.DispersiveUtils)
        from molass_legacy.Models.Stochastic.DispersiveUtils import investigate_sdm_params_from_optimizer_params
        print("show_sdm_investigator")
        optimizer = self.fullopt
        i = self.get_best_index()
        x_array = self.demo_info[1]
        p = x_array[i,:]
        investigate_sdm_params_from_optimizer_params(optimizer, p)

    def show_sdm_lrf_investigator(self, debug=True):
        if debug:
            import Models.Stochastic.DispersiveUtils
            from importlib import reload
            reload(Models.Stochastic.DispersiveUtils)
        from molass_legacy.Models.Stochastic.DispersiveUtils import investigate_sdm_lrf_params_from_optimizer_params
        print("show_sdm_lrf_investigator")
        optimizer = self.fullopt
        i = self.get_best_index()
        x_array = self.demo_info[1]
        p = x_array[i,:]
        investigate_sdm_lrf_params_from_optimizer_params(optimizer, p)

    def show_adhoc_figure(self, debug=True):
        if debug:
            import Optimizer.AdhocFigure
            from importlib import reload
            reload(Optimizer.AdhocFigure)
        from molass_legacy.Optimizer.AdhocFigure import show_adhoc_figure_impl
        show_adhoc_figure_impl(self)
    
    def change_view_widths(self):
        from molass_legacy._MOLASS.SerialSettings import set_setting
        from molass_legacy.Trimming.OptViewRange import OptViewRange
        if self.optview_ranges is None:
            optimizer = self.fullopt
            ranges = []
            for curve in [optimizer.uv_curve, optimizer.xr_curve]:
                x = curve.x
                y = curve.y
                range_ = OptViewRange(x, y, upper=0.03)
                ranges.append(range_.get_range())
        else:
            ranges = None

        self.optview_ranges = ranges
        self.set_view_widths(ranges=ranges)
        self.fig.canvas.draw()
        self.update_view_width_button()
    
    def set_view_widths(self, ranges=None):
        if self.initial_view_ranges is None:
            ranges = []
            for ax in self.axes[0:2]:
                ranges.append(ax.get_xlim())
            self.initial_view_ranges = ranges
 
        if ranges is None:
            ranges = self.initial_view_ranges
        
        for ax, range_ in zip(self.axes[0:2], ranges):
            ax.set_xlim(range_)
    
    def show_this_figure(self):
        from importlib import reload
        import Optimizer.XrStateFigure
        reload(Optimizer.XrStateFigure)
        from molass_legacy.Optimizer.XrStateFigure import show_this_figure_impl
        show_this_figure_impl(self)




# coding: utf-8
"""
    UnifiedModel.py

    Copyright (c) 2020, SAXS Team, KEK-PF
"""
import numpy as np
from .GaussianMixture import gm_curve
from molass_legacy.KekLib.OurMatplotlib import get_color
import molass_legacy.KekLib.DebugPlot as plt

SHOW_PROPORTIONS = False

def tutorial():

    if SHOW_PROPORTIONS:
        nrows = 2
        fighight = 10
    else:
        nrows = 1
        fighight = 5

    fig, axes = plt.subplots(nrows=nrows, ncols=3, figsize=(18, fighight))
    fig.suptitle("Unified Modelling of SEC-SAXS", fontsize=20)

    u_title_list = [
            "Source Distribution caused from Rg",
            "Apparant Distribution in Absorbance depending on Wavelength",
            "Apparant Distribution in Scattering depending on Angle"
            ]
    d_title_list = [
            "Component Proportions caused from Rg",
            "Component Proportions in Absorbance",
            "Component Proportions in Scattering"
            ]

    if SHOW_PROPORTIONS:
        uaxes = axes[0,:]
        daxes = axes[1,:]
        title_y = 1.02
    else:
        uaxes = axes
        daxes = [None]*3
        title_y = 1.01

    for axu, u_title, axd, u_title in zip(uaxes, u_title_list, daxes, d_title_list):
        axu.set_title(u_title, fontsize=14, y=title_y)
        if axd:
            axd.set_title(u_title, fontsize=14, y=title_y)

    x_length_list = [100, 600, 300]

    params_list = [
            [(100/3, 30/3, 1), (200/3, 20/3, 1)],
            [(200, 60, 0.7), (400, 40, 1.2)],
            [(100, 30, 1.2), (200, 20, 0.5)],
            ]

    min_ymin = np.inf
    max_ymax = -np.inf
    for axu, axd, length, params in zip(uaxes, daxes, x_length_list, params_list):
        x, y, gy = gm_curve(x_length=length, params=params, ret_detail=True)
        scale = length/400
        axu.plot(x, y*scale)
        for k, gy_ in enumerate(gy):
            axu.plot(x, gy_*scale, ':')
            if axd:
                axd.plot(x, gy_/y, ':', color=get_color(k+1))

        ymin, ymax = axu.get_ylim()
        min_ymin = min(min_ymin, ymin)
        max_ymax = max(max_ymax, ymax)

    for axu in uaxes:
        axu.set_ylim(min_ymin, max_ymax)

    fig.tight_layout()
    fig.subplots_adjust(top=0.91 if SHOW_PROPORTIONS else 0.85)

    plt.show()



# coding: utf-8
"""
    Rgg.SpikeData.py

    Copyright (c) 2021, SAXS Team, KEK-PF
"""
import numpy as np
from molass_legacy.Peaks.ElutionModels import egh
from .RggUtils import normal_pdf

FULL_PARAMS =  np.array([
        [35, [1.0, 200, 20,  0]],
        [31, [0.2, 250, 20,  0]],
        [23, [0.8, 400, 30,  0]],
        ])

def generate_demo_data(num_components=3):
    x = np.arange(600)
    y = np.zeros(len(x))
    rg_list = []
    cy_list = []

    gen_params = FULL_PARAMS if num_components == 3 else FULL_PARAMS[[0,2]]
    for rg, params in gen_params:
        if False:
            cy = egh(x, *params)
        else:
            h, mu, sigma, tau = params
            cy = h*normal_pdf(x, mu, sigma)
        cy_list.append(cy)
        y += cy
        rg_list.append(rg)

    rg = np.sum(np.array(rg_list)[:,np.newaxis]*np.array(cy_list)/y, axis=0)
    return x, y, rg, len(cy_list)



"""
    Optimizer.TwoParamAnalysis.py

    Copyright (c) 2021-2022, SAXS Team, KEK-PF
"""
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from molass_legacy.KekLib.OurTkinter import Tk, Dialog, ToolTip
from molass_legacy.KekLib.OurMatplotlib import NavigationToolbar
from molass_legacy._MOLASS.SerialSettings import get_setting
from ttkwidgets import TickScale
from DataUtils import get_in_folder

INIT_ZOOM_ARE_RATIO = 0.5
MIN_ZOOM_ARE_RATIO = 0.1
MAX_ZOOM_ARE_RATIO = 1.0
ZOOM_AREA_RESOLUTION = 0.01

class TwoParamAnalysis(Dialog):
    def __init__(self, parent, dsets, optimizer, demo_info, curr_index, i=5, j=9):
        self.parent = parent
        self.in_folder = get_in_folder()
        self.dsets = dsets
        self.optimizer = optimizer
        self.xarray = demo_info[1]
        self.params = self.xarray[curr_index]
        (i, j), spp = self.select_default_params()
        self.i = i
        self.j = j
        self.curr_index = curr_index
        self.set_trace_xy()
        self.spp = spp
        self.temp_params = self.params.copy()
        self.bounds = optimizer.real_bounds
        self.two_param_vector_func = np.vectorize(self.two_param_func)
        self.zoom_area_ratio = None     # this will be replaced by TickScale
        Dialog.__init__(self, parent, "Two Parameter Analysis", visible=False)

    def set_trace_xy(self):
        self.trace_xy = self.xarray[0:self.curr_index+1,[self.i,self.j]].T

    def select_default_params(self):
        xr_params, xr_baseparams, rgs, mapping, uv_params, uv_baseparams, mappable_range, seccol_params = self.optimizer.split_params_simple(self.params)[0:8]

        def normalize(p):
            a = np.average(p)
            s = np.std(p)
            return (p - a)/s

        if len(xr_params.shape) == 2:
            xr_params_ = xr_params[:,0]
        else:
            xr_params_ = xr_params
        scales = normalize(xr_params_) + normalize(uv_params[:])
        k = len(scales)-2
        pp = np.argpartition(scales, k)
        # print("scales=", scales, "k=", k, "pp=", pp)
        spp = sorted(pp[k:])
        if len(xr_params.shape) == 2:
            # major two peak positions
            i, j = np.arange(1, len(xr_params.flatten()), 4)[spp]
        else:
            # major two rgs
            i, j = len(xr_params) + len(xr_baseparams) + np.arange(0, len(rgs))[spp]
        return (i, j), spp

    def two_param_func(self, x, y):
        self.temp_params[self.i] = x
        self.temp_params[self.j] = y
        return self.optimizer.objective_func(self.temp_params)

    def show(self):
        self._show()

    def cancel(self):
        plt.close(self.fig)
        Dialog.cancel(self)

    def body(self, body_frame):
        cframe = Tk.Frame(body_frame)
        cframe.pack()
        tframe = Tk.Frame(body_frame)
        tframe.pack(fill=Tk.X, padx=20)
        tframe_left = Tk.Frame(tframe)
        tframe_left.pack(side=Tk.LEFT)
        tframe_right = Tk.Frame(tframe)
        tframe_right.pack(side=Tk.RIGHT)

        self.fig = fig = plt.figure(figsize=(21, 8))
        self.fig_suptitle = fig.suptitle(self.get_suptitle_text(), fontsize=20)
        ax1 = fig.add_subplot(131, projection="3d")
        ax2 = fig.add_subplot(132)
        ax3 = fig.add_subplot(133)
        self.axes = [ax1, ax2, ax3]
        self.draw_function()
        fig.tight_layout()
        fig.subplots_adjust(top=0.88, left=0.01, wspace=0.18, right=0.99)
        self.mpl_canvas = FigureCanvasTkAgg(self.fig, cframe)
        self.mpl_canvas_widget = self.mpl_canvas.get_tk_widget()
        self.mpl_canvas_widget.pack(fill=Tk.BOTH, expand=1)
        self.toolbar = NavigationToolbar(self.mpl_canvas, tframe_left)
        self.toolbar.update()
        self.build_control_frame(tframe_right)

    def build_control_frame(self, frame):
        inspect_btn = Tk.Button(frame, text="Parameter Inspection", command=self.show_params_inspect)
        inspect_btn.pack(side=Tk.LEFT, padx=20)
        change_btn = Tk.Button(frame, text="Change Parameters", command=self.show_params_selector)
        change_btn.pack(side=Tk.LEFT, padx=20)

        ratio_frame = Tk.Frame(frame)
        ratio_frame.pack(side=Tk.LEFT, padx=20)
        label = Tk.Label(ratio_frame, text="Zoom Area Ratio: ")
        label.pack(side=Tk.LEFT)
        self.zoom_area_ratio = TickScale(ratio_frame, orient='horizontal', from_=0.1, to=1.0, resolution=ZOOM_AREA_RESOLUTION)
        self.zoom_area_ratio.pack(side=Tk.LEFT)
        self.zoom_area_ratio.set(INIT_ZOOM_ARE_RATIO)
        ToolTip(ratio_frame, 'Use arrow keys (←, →) to adjust this ratio.')

        redraw_btn = Tk.Button(frame, text="Redraw", command=self.redraw)
        redraw_btn.pack(side=Tk.LEFT, padx=20)

        self.bind("<Left>", self.on_arrow_touch)
        self.bind("<Right>", self.on_arrow_touch)
        self.state_sensitive_widgets = [inspect_btn, change_btn, self.zoom_area_ratio, redraw_btn]

    def on_arrow_touch(self, event):
        if event.keysym == "Left":
            d = -ZOOM_AREA_RESOLUTION
        else:
            d = ZOOM_AREA_RESOLUTION
        new_value = max(MIN_ZOOM_ARE_RATIO, min(MAX_ZOOM_ARE_RATIO, self.zoom_area_ratio.get() + d))
        self.zoom_area_ratio.set(new_value)

    def get_zoom_area_ratio(self):
        return INIT_ZOOM_ARE_RATIO if self.zoom_area_ratio is None else self.zoom_area_ratio.get()

    def buttonbox(self):
        box = Tk.Frame(self)
        box.pack()

        w = Tk.Button(box, text="◀ Close", width=10, command=self.cancel)
        w.pack(pady=10)

        self.bind("<Escape>", self.cancel)

    def get_area(self, zoom=None):
        xmin, xmax = self.bounds[self.i]
        ymin, ymax = self.bounds[self.j]

        if zoom is None:
            xmin_, xmax_ = xmin, xmax
            ymin_, ymax_ = ymin, ymax
        else:
            cx, cy, proportion = zoom
            dx = (xmax - xmin)*proportion/2
            dy = (ymax - ymin)*proportion/2
            xmin_, xmax_ = cx - dx, cx + dx
            ymin_, ymax_ = cy - dy, cy + dy

        return xmin_, xmax_, ymin_, ymax_

    def compute_surface(self, xmin, xmax, ymin, ymax):
        x = np.linspace(xmin, xmax, 40)
        y = np.linspace(ymin, ymax, 40)
        xx, yy = np.meshgrid(x, y)
        zz = self.two_param_vector_func(xx, yy)
        return xx, yy, zz

    def get_suptitle_text(self):
        return "Two Parameter Analysis for %s with selected parameters(%d, %d)" % (self.in_folder, self.i, self.j)

    def draw_function(self):
        ax1, ax2, ax3 = self.axes
        ax1.set_title("3D View (Zoomed)", fontsize=16)
        ax2.set_title("Contour View (Whole)", fontsize=16)
        ax3.set_title("Contour View (Zoomed)", fontsize=16)

        trace_color = "green"
        trace_kwargs = {"color":trace_color, "label":"minima trace", "alpha":0.5}
        initial_kwargs = {"color":"cyan", "label":"initial position"}
        current_kwargs = {"color":"yellow", "label":"current position"}
        # cmap = "viridis"
        cmap = "copper"

        cx = self.params[self.i]
        cy = self.params[self.j]
        warea = self.get_area()
        wxx, wyy, wzz = self.compute_surface(*warea)
        ratio = self.get_zoom_area_ratio()
        sarea = self.get_area(zoom=(cx, cy, ratio))
        sxx, syy, szz = self.compute_surface(*sarea)

        xlabel = "parameter %d" % self.i
        ylabel = "parameter %d" % self.j
        ax1.plot_surface(sxx, syy, szz, cmap=cmap)
        ax1.set_xlabel(xlabel)
        ax1.set_ylabel(ylabel)
        ax1.set_zlabel("FV")

        tx, ty = self.trace_xy
        tz = self.two_param_vector_func(tx, ty)
        cz = self.two_param_func(cx, cy)
        ax1.plot(tx[0], ty[0], tz[0], "o", **initial_kwargs)
        ax1.plot(tx, ty, tz, **trace_kwargs)
        ax1.plot(cx, cy, cz, "o", **current_kwargs)

        levels = None
        for ax, xx, yy, zz in [(ax3, sxx, syy, szz), (ax2, wxx, wyy, wzz)]:
            ax.set_xlabel(xlabel)
            ax.set_ylabel(ylabel)
            cs = ax.contourf(xx, yy, zz, levels=levels, cmap=cmap)
            levels = cs.levels
            xmin, xmax = ax.get_xlim()
            ax.set_xlim(xmin, xmax)
            ymin, ymax = ax.get_ylim()
            ax.set_ylim(ymin, ymax)
            ax.plot([cx, cx], [ymin, ymax], color="yellow")
            ax.plot([xmin, xmax], [cy, cy], color="yellow")
            ax.plot(tx[0], ty[0], "o", **initial_kwargs)
            ax.plot(tx, ty, **trace_kwargs)
            ax.plot(cx, cy, "o", **current_kwargs)

        xmin_, xmax_, ymin_, ymax_ = sarea
        x = [xmin_, xmax_, xmax_, xmin_, xmin_]
        y = [ymin_, ymin_, ymax_, ymax_, ymin_]
        ax2.plot(x, y, ":", color="red", label="zoomed area")

        for ax in [ax1, ax2, ax3]:
            ax.legend()

    def change_state(self, state):
        for w in self.state_sensitive_widgets:
            w.config(state=state)

    def redraw(self):
        self.change_state(Tk.DISABLED)
        self.update()
        self.set_trace_xy()
        self.fig_suptitle.set_text(self.get_suptitle_text())
        for ax in self.axes:
            ax.cla()

        self.draw_function()
        self.mpl_canvas.draw()
        self.change_state(Tk.NORMAL)

    def show_params_inspect(self):
        from .ParamsInspection import ParamsInspection
        dialog = ParamsInspection(self.parent, self.params, self.dsets, self.optimizer)
        dialog.show()

    def show_params_selector(self):
        from .ParamsSelector import ParamsSelector
        dialog = ParamsSelector(self, self.optimizer, np.arange(len(self.params)))
        dialog.set_selection([self.i, self.j])
        dialog.show()
        selection = dialog.get_selection()
        if selection is None:
            return
        else:
            self.i, self.j = selection
            self.redraw()



"""
    SecTheory.EghCompat.py

    Copyright (c) 2022-2024, SAXS Team, KEK-PF
"""
import logging
import numpy as np
import SecTheory.MonoPore
from importlib import reload
reload(SecTheory.MonoPore)
from SecTheory.MonoPore import estimate_monopore_params, get_modified_params
import SecTheory.T0UpperBound
reload(SecTheory.T0UpperBound)
from SecTheory.T0UpperBound import estimate_t0upper_bound
import SecTheory.MomentAnalysis
reload(SecTheory.MomentAnalysis)

def verify_basic_assumptions(prep_info, nc=5, first_only=False, global_opt=False):

    logger = logging.getLogger(__name__)

    sd = prep_info.sd
    in_folder = prep_info.in_folder
    rg_curve = prep_info.rg_curve

    D, E, qv, ecurve = sd.get_xr_data_separate_ly()

    ecurve = sd.get_xray_curve()
    # estimate_sec_params(ecurve, debug=True)
    # return

    t0_upper_bound = estimate_t0upper_bound(ecurve)
    print("------------------- t0_upper_bound=", t0_upper_bound)
    ret = estimate_monopore_params(ecurve, rg_curve, nc,
                                    t0_upper_bound=t0_upper_bound, global_opt=global_opt, logger=logger, debug=True)
    if first_only:
        return

    modified_params = get_modified_params(ecurve, nc, ret.x, logger=logger, debug=True)
    if modified_params is not None:
        ret = estimate_monopore_params(ecurve, rg_curve, nc,
                                        t0_upper_bound=t0_upper_bound, init_params=modified_params, logger=logger, debug=True)



"""
    CrysolUtils.py

    Copyright (c) 2021-2023, SAXS Team, KEK-PF
"""
import os
import re
import numpy        as np

def np_loadtxt_crysol( filename, encoding='cp932' ):
    fh = open( filename, encoding=encoding )

    comment_lines = []

    def generator( fh ):
        for line in fh:
            if len( comment_lines ) < 1:
                comment_lines.append( line )
            else:
                yield line

    try:
        array = np.loadtxt( generator( fh ) )
    except:
        from molass_legacy.KekLib.ExceptionTracebacker import log_exception
        log_exception(None, "np_loadtxt_crysol: ")
        array = None

    fh.close()
    return array, comment_lines



# coding: utf-8
"""
    PhaseRetrievalDemo.py

    Copyright (c) 2020, SAXS Team, KEK-PF
"""
import copy
import time
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
import matplotlib.cm as cm
from mpl_toolkits.mplot3d import Axes3D
from .PhaseRetrieval import plot_3d_image, plot_3d_scatter, fienup_phase_retrieval
from .ReciprocalData import ReciprocalData
from .SaxsSamples import BallVoxels

def ball_demo(**kwargs):
    from .SaxsSamples import BallVoxels
    n = kwargs.get('n')
    radius = kwargs.pop('radius', 0.3)
    ball = BallVoxels(shape=(n,n,n), radius=radius)
    object_demo(ball, **kwargs)

def disc_demo(**kwargs):
    from .SaxsSamples import DiscVoxels
    n = kwargs.get('n')
    radius = kwargs.pop('radius', 0.5)
    height = kwargs.pop('height', 0.2)
    disc = DiscVoxels(shape=(n,n,n), radius=radius, height=height)
    object_demo(disc, **kwargs)

def torus_demo(**kwargs):
    from .SaxsSamples import TorusVoxels
    n = kwargs.get('n')
    R = kwargs.pop('R', 0.3)
    r = kwargs.pop('r', 0.1)
    torus = TorusVoxels(shape=(n,n,n), R=R, r=r)
    object_demo(torus, **kwargs)

def object_demo(obj_voxels, fig=None, n=32, gpu=None, exact_mag=False, use_mask=False):
    if fig is None:
        fig_ = plt.figure(figsize=(24,15))
    else:
        fig_ = fig

    if gpu is None:
        from Env.EnvInfo import get_global_env_info
        env_info = get_global_env_info()
        gpu = env_info.nvidiagpu_is_available

    t0 = time.time()

    density = 0.1
    data = obj_voxels.get_data(density=density)
    q = np.linspace(0, 0.5, 600)
    rdata = ReciprocalData(data.shape)
    F = rdata.get_reciprocal(data)
    y = rdata.get_scattering_curve(q, F)
    x = rdata.qbinsc
    # print('rdata.qbin_labels=', rdata.qbin_labels)
    print('rdata.qbin_labels.dtype=', rdata.qbin_labels.dtype)
    qv = rdata.interp(x)
    absF = np.abs(F)
    if exact_mag:
        magnitudes = absF
    else:
        magnitudes = rdata.get_3d_magnitudes(qv)
    min_value = density/2
    if use_mask:
        ball = BallVoxels(shape=(n,n,n), radius=n)
        mask = ball.get_data()
    else:
        mask = None
    result = fienup_phase_retrieval(magnitudes, mask=mask,
                            steps=10000, gpu=gpu,
                            min_value=min_value,
                            verbose=True)

    print('It took %.3g seconds.' % (time.time() - t0))

    gs = gridspec.GridSpec(3,4)
    fig_.suptitle("Phase Retrieval for 3D in SAXS with (%d,%d,%d) voxels" % (n,n,n), fontsize=20)
    ax0 = fig_.add_subplot(gs[0,0], projection='3d')
    ax1 = fig_.add_subplot(gs[0,1], projection='3d')
    ax2 = fig_.add_subplot(gs[0,2], projection='3d')
    ax3 = fig_.add_subplot(gs[0,3])
    ax4 = fig_.add_subplot(gs[1,0], projection='3d')
    ax5 = fig_.add_subplot(gs[1,1], projection='3d')
    ax6 = fig_.add_subplot(gs[1,2], projection='3d')
    ax7 = fig_.add_subplot(gs[1,3])
    ax8 = fig_.add_subplot(gs[2,0], projection='3d')
    ax9 = fig_.add_subplot(gs[2,1], projection='3d')
    axa = fig_.add_subplot(gs[2,2], projection='3d')
    axb = fig_.add_subplot(gs[2,3])

    fontsize = 16
    ax0.set_title('Time Domain Object', y=1.1, fontsize=fontsize)
    ax1.set_title('Frequency Domain abs(F) (from Object)', y=1.1, fontsize=fontsize)
    ax2.set_title('qbin_labels for 3D → 1D spherical average', y=1.1, fontsize=fontsize)
    ax3.set_title('Scattering Curve', fontsize=fontsize)
    ax4.set_title('Time Domain Retrieved Object', y=1.1, fontsize=fontsize)
    ax5.set_title('Frequency Domain abs(F) (from Curve)', y=1.1, fontsize=fontsize)
    ax6.set_title('abs(F)(from Curve) - abs(F)(from Ball)', y=1.1, fontsize=fontsize)
    ax7.set_title('qbin values (adopted points for phase retrieval)', fontsize=fontsize)
    ax8.set_title('Time Domain Retrieved Object', y=1.1, fontsize=fontsize)
    ax9.set_title('Frequency Domain abs(F) (from Retrieved Object)', y=1.1, fontsize=fontsize)
    axa.set_title('qbin_labels for 3D → 1D spherical average', y=1.1, fontsize=fontsize)
    axb.set_title('Scattering Curve from Retrieved Object', fontsize=fontsize)

    imean = rdata.Imean[rdata.Imean > 0]
    ymax = np.log10(imean.max())
    ymin = np.log10(imean.min())
    ymax_ = ymin*(-0.1) + ymax*1.1
    ymin_ = ymin*1.1 + ymax*(-0.1)
    for ax in [ax3, ax7, axb]:
        ax.set_ylim(ymin_, ymax_)
        ax.set_xlabel('Q')
        ax.set_ylabel('$Log_{10}$(Intensity)')

    plot_3d_scatter(ax0, data, shape_limits=True)
    logy = np.log10(y)
    ax3.plot(q, logy, color='orange')
    log_imean = np.log10(rdata.Imean)
    ax3.scatter(x, log_imean, cmap=cm.cool, c=x)

    plot_3d_scatter(ax2, rdata.qbin_labels, cmap=cm.cool)
    plot_3d_scatter(ax1, absF, cmap=cm.bwr)

    ax7.plot(q, logy, color='orange', label='simulated curve')
    ax7.scatter(x, np.log10(qv), cmap=cm.cool, c=x)
    yr = rdata.get_scattering_curve(q, magnitudes, absolute=True)
    ax7.plot(q, np.log10(yr), ':', color='green', label='curve for verification')
    ax7.legend()

    diff = magnitudes - absF
    plot_3d_scatter(ax6, diff, cmap=cm.bwr)
    plot_3d_scatter(ax5, magnitudes, cmap=cm.bwr)
    plot_3d_scatter(ax4, result, min_value=min_value, shape_limits=True)

    data_verif = copy.deepcopy(result)
    data_verif[data_verif < min_value] = 0
    plot_3d_scatter(ax8, data_verif, shape_limits=True)
    Fr = rdata.get_reciprocal(data_verif)
    absFr = np.abs(Fr)
    plot_3d_scatter(ax9, absFr, cmap=cm.bwr)
    plot_3d_scatter(axa, rdata.qbin_labels, cmap=cm.cool)

    yv = rdata.get_scattering_curve(q, Fr)
    axb.plot(q, logy, color='orange', label='simulated curve')
    axb.plot(q, np.log10(yv), ':', color='blue', label='curve from retrieved object')
    axb.set_xlim(ax7.get_xlim())
    axb.set_ylim(ax7.get_ylim())
    axb.legend()

    fig_.tight_layout()
    fig_.subplots_adjust(top=0.9)

    if fig is None:
        plt.show()



"""
    Models.Stochastic.LnporeUtils.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import numpy as np

from molass_legacy.Models.Stochastic.LognormalUtils import compute_mode

def compute_lnp_elutions(x, lnp_params, rgs):
    from molass_legacy.Models.Stochastic.LognormalPoreFunc import lognormal_pore_func
    N, T, x0, me, mp, mu, sigma = lnp_params[0:7]
    scales = lnp_params[7:]
    cy_list = []
    for scale, rg in zip(scales, rgs):
        cy = lognormal_pore_func(x, scale, N, T, me, mp, mu, sigma, rg, x0) 
        cy_list.append(cy)
    ty = np.sum(cy_list, axis=0)
    return cy_list, ty

def plot_elution_exclusion_impl(ax1, x, y, lnp_params, rgs):

    from molass_legacy.Models.Stochastic.MonoporeUtils import draw_exclusion_cuve
    cy_list, ty = compute_lnp_elutions(x,lnp_params, rgs)
    ax1.set_title("Elution Curve with Exclusion Curve", fontsize=16)
    ax1.set_xlabel("Time (frames)")
    ax1.set_ylabel("Intensity")
    ax1.plot(x, y, color='orange',label="data")
    for k, (cy, rg) in enumerate(zip(cy_list, rgs)):
        ax1.plot(x, cy, ':', label="component-%d, Rg=%.3g" % (k, rg))
    ax1.plot(x, ty, ':', color="red", label="model total")
    ax1.legend()
    axt = ax1.twinx()
    axt.grid(False)
    trs = [x[0] + np.argmax(cy) for cy in cy_list]
    N, T, x0, me, mp, mu, sigma = lnp_params[0:7]
    mode = compute_mode(mu, sigma)
    draw_exclusion_cuve(axt, (N, T, x0, me, mp, mode), trs, rgs)

def compute_psd_curve(mu, sigma):
    from molass_legacy.Models.Stochastic.ParamLimits import PORESIZE_INTEG_LIMIT
    from molass_legacy.Models.Stochastic.LognormalPoreFunc import distr_func
    pv = np.linspace(0, PORESIZE_INTEG_LIMIT, 100)
    pd = distr_func(pv, mu, sigma)
    return pv, pd

def plot_lognormal_psd_impl(ax2, mu, sigma, return_artists=False):
    mode = compute_mode(mu, sigma)
    ax2.set_title("Lognormal Pore Size Distribution", fontsize=16)
    ax2.set_xlabel(r"Pore Size ($\AA$)")
    ax2.set_ylabel("Density")
    pv, pd = compute_psd_curve(mu, sigma)
    psd_curve, = ax2.plot(pv, pd, label="lognormal pdf")
    psd_mode = ax2.axvline(x=mode, ls=':', color="blue", label="poresize (%d)" % int(mode))
    ax2.legend()
    if return_artists:
        from molass_legacy.KekLib.MatplotlibUtils import get_labeltext_from_line
        artists = psd_curve, psd_mode
        labels = [get_labeltext_from_line(a) for a in artists]
        return *artists, *labels

def plot_lnpore_component_with_sliders_impl(x, y, lnp_params, rgs, title=None):
    from matplotlib.gridspec import GridSpec
    from matplotlib.widgets import Slider, Button
    import molass_legacy.KekLib.DebugPlot as plt
    from molass_legacy.SerialAnalyzer.DataUtils import get_in_folder
    from molass_legacy.Models.Stochastic.LognormalUtils import compute_mode, compute_stdev, compute_mu_sigma

    plot_params = lnp_params.copy()
    mu, sigma = lnp_params[5:7]
    mode = compute_mode(mu, sigma)
    stdev = compute_stdev(mu, sigma)
    plot_params[5:7] = mode, stdev
    slider_specs = [    ("N", 0, 6000, plot_params[0]),
                        ("T", 0, 3, plot_params[1]),
                        ("t0",  -1000, 1000, plot_params[2]),
                        ("me", 0, 3, plot_params[3]),
                        ("mp", 0, 3, plot_params[4]),
                        ("mode", 20, 400, plot_params[5]),
                        ("stdev", 1, 100, plot_params[6]),
                        ]

    with plt.Dp(button_spec=["OK", "Cancel"]):
        fig = plt.figure(figsize=(18,5))
        gs = GridSpec(1,5)
        ax1 = fig.add_subplot(gs[0,0:3])
        ax2 = fig.add_subplot(gs[0,3:5])
        if title is None:
            title = "Lognormal PSD Elution Components for %s" % get_in_folder()
        fig.suptitle(title, fontsize=20, x=0.35)
        ax1.set_title("Component Elution Curves", fontsize=16)
        ax1.plot(x, y, color='orange',label="data")
        cy_list, ty = compute_lnp_elutions(x, lnp_params, rgs)
        ecurves = []
        for k, (cy, rg) in enumerate(zip(cy_list, rgs)):
            curve, = ax1.plot(x, cy, ':', label="component-%d, Rg=%.3g" % (k, rg))
            ecurves.append(curve)
        curve, = ax1.plot(x, ty, ':', color="red", label="model total")
        ecurves.append(curve)
        ax1.legend()
        psd_curve, psd_mode, psd_curve_label, psd_mode_label = plot_lognormal_psd_impl(ax2, mu, sigma, return_artists=True)

        data_ymin = y.min()
        data_ymax = y.max()
        w = -0.05
        ax1_ymin = data_ymin*(1 - w) + data_ymax*w
        w = 1.1
        ax1_ymax = data_ymin*(1 - w) + data_ymax*w

        def slider_update(k, val):
            # print([k], "slider_update", val)
            plot_params[k] = val
            mode, stdev = plot_params[5:7]
            mu, sigma = compute_mu_sigma(mode, stdev)
            pv, pd = compute_psd_curve(mu, sigma)
            psd_curve.set_ydata(pd)
            psd_mode.set_xdata(mode)
            psd_mode_label.set_text("poresize (%d)" % int(mode))
            lnp_params_temp = plot_params.copy()
            lnp_params_temp[5:7] = mu, sigma
            cy_list, ty = compute_lnp_elutions(x, lnp_params_temp, rgs)
            for cy, ecurve in zip(cy_list + [ty], ecurves):
                ecurve.set_ydata(cy)
            tymin = ty.min()
            tymax = ty.max()
            ax1.set_ylim(min(ax1_ymin, tymin), max(ax1_ymax, tymax))
            ymin, ymax = ax2.get_ylim()
            ax2.set_ylim(ymin, pd.max()*1.2)
            fig.canvas.draw_idle()

        slider_axes = []
        sliders = []
        for k, (label, valmin, valmax, valinit) in enumerate(slider_specs, start=0):
            ax_ = fig.add_axes([0.75, 0.8 - 0.08*k, 0.2, 0.03])
            slider  = Slider(ax_, label=label, valmin=valmin, valmax=valmax, valinit=valinit)
            slider.on_changed(lambda val, k_=k: slider_update(k_, val))
            slider_axes.append(ax_)
            sliders.append(slider)

        def reset(event):
            print("reset")
            for k, slider in enumerate(sliders):
                slider.reset()

        button_ax = fig.add_axes([0.85, 0.2, 0.12, 0.05])
        debug_btn = Button(button_ax, 'Reset', hovercolor='0.975')
        debug_btn.on_clicked(reset)

        fig.tight_layout()
        fig.subplots_adjust(right=0.7)
        ret = plt.show()
    return ret


"""
    CheckEnvDialog.py

    Copyright (c) 2019-2022, SAXS Team, KEK-PF
"""
from molass_legacy.KekLib.OurTkinter import Tk, Dialog
from molass_legacy.KekLib.ReadOnlyText import ReadOnlyText
from .EnvMessages import get_env_messages

class CheckEnvDialog(Dialog):
    def __init__(self, parent):
        self.grab = 'local'     # used in grab_set
        self.parent = parent
        Dialog.__init__( self, parent, "Environment Info", visible=False )

    def show(self):
        self._show()

    def body(self, body_frame):
        iframe = Tk.Frame(body_frame)
        iframe.pack(side=Tk.LEFT, padx=10, pady=5)

        self.info_text = ReadOnlyText(iframe, width=80, height=10)
        self.info_text.pack()

        for line in get_env_messages():
            self.info_text.insert(Tk.END, line)

    def buttonbox( self ):
        box = Tk.Frame(self)
        box.pack()

        self.ok_btn = Tk.Button(box, text="OK", width=10, command=self.ok, default=Tk.ACTIVE)
        self.ok_btn.pack(side=Tk.LEFT, padx=5, pady=5)

        self.bind("<Return>", self.ok)
        self.bind("<Escape>", self.cancel)



"""
    main.py

    Copyright (c) 2016-2025, SAXS Team, KEK-PF
"""

numba_is_available = False
DEBUG = False

def molass_gui():
    """
    molass_gui()
    This function initializes the Molass GUI application.
    It sets up the environment, imports necessary modules, and starts the main GUI loop.
    It also handles debugging and logging if DEBUG is set to True.
    It is designed to be run as the main entry point of the application.

    Note that the entry point is defined in pyproject.toml as follows::

        [project.scripts]
        molass = "molass_legacy.main:molass_gui"

    This definition allows the application to be run from the command line using the `molass` command.
    
    """
    global print
    if DEBUG:
        import logging
        logging.basicConfig(level=logging.INFO,
                            format='%(asctime)s %(message)s',
                            datefmt='%Y-%m-%d %H:%M:%S')
        logger = logging.getLogger()
        logger.addHandler(logging.FileHandler('molass-debug.log', 'a'))
        print = logger.info

    global numba_is_available

    print( 'import strftime' )
    from time   import strftime
    print( strftime('%H:%M:%S import sys') )
    import sys

    print( strftime('%H:%M:%S import os') )
    import os

    print( strftime('%H:%M:%S tkinter') )
    import tkinter as Tk
    import tkinter.ttk as ttk

    try:
        print( strftime('%H:%M:%S screeninfo') )
        import screeninfo
    except:
        print( strftime('%H:%M:%S failed to import screeninfo') )

    # ---- sys.path begin --------------------------------------------------
    this_dir = os.path.abspath(os.getcwd())
    lib_dir = os.path.dirname(this_dir)

    assert os.path.exists(lib_dir)

    sys.path.insert(0, lib_dir)
    # ---- sys.path end ----------------------------------------------------

    import win32console, win32con
    import psutil

    import molass_legacy.KekLib
    from molass_legacy.KekLib.Console    import Console
    console = Console( windows_app_only=True, console_parent_only=True )

    from molass_legacy.KekLib.TkUtils import get_tk_root
    from molass_legacy._MOLASS.SplashMessage  import AppSplashScreen
    #-------------------------------------------------------------------------------
    #   GUI
    #-------------------------------------------------------------------------------
    root = get_tk_root(withdraw=False)  # using get_tk_root to set DebugPlot environment
    splash = AppSplashScreen( root, 26, 3 )

    # ----- pyinstaller help imports begin -----------------------------------------
    print( strftime('%H:%M:%S tkinter other modules') )
    import tkinter.simpledialog
    import tkinter.scrolledtext
    import tkinter.filedialog
    import tkinter.font
    # import tkinter.tix
    splash.update()

    # print( strftime('%H:%M:%S pyautogui') )
    # import pyautogui
    # splash.update()
    print( strftime('%H:%M:%S packaging.version') )
    import packaging.version
    splash.update()
    print( strftime('%H:%M:%S packaging.specifiers') )
    import packaging.specifiers
    splash.update()
    print( strftime('%H:%M:%S numpy') )
    import numpy
    splash.update()
    #    mkl_intel_thread.dll
    #    mkl_core.dll
    #    libiomp5md.dll
    #    mkl_avx2.dll
    #    mkl_def.dll
    print( strftime('%H:%M:%S pylab') )
    import pylab
    splash.update()
    # ----- pyinstaller help imports end   -----------------------------------------
    print( strftime('%H:%M:%S glob') )
    import glob
    splash.update()
    print( strftime('%H:%M:%S scipy.stats') )
    import scipy.stats
    splash.update()
    print( strftime('%H:%M:%S lmfit') )
    import lmfit
    splash.update()
    print( strftime('%H:%M:%S statsmodels.api') )
    import statsmodels.api
    splash.update()
    print( strftime('%H:%M:%S seaborn') )
    import seaborn
    splash.update()
    print( strftime('%H:%M:%S sklearn') )
    from sklearn import gaussian_process
    splash.update()
    print( strftime('%H:%M:%S openpyxl') )
    import openpyxl
    splash.update()

    try:
        print( strftime('%H:%M:%S win32com.client') )
        import win32com.client
        splash.update()
    except PermissionError:
        print("PermissionError: failed to import win32com.client")
        print("\nIf you have installed Python as administrator,")
        print("you need to run molass once as administrator due to pywin32 permission issues.")
        print("After that, you should be able to run molass as a normal user.")
        exit(-1)

    print( strftime('%H:%M:%S checking pywin32_postinstall execution') )
    from molass.PackageUtils.PyWin32Utils import check_pywin32_postinstall
    if not check_pywin32_postinstall():
        print("Please run (possibly as administrator) the following command to fix the issue:")
        print("python -m pywin32_postinstall -install")
        exit(-1)

    print( strftime('%H:%M:%S win32process') )
    import win32process
    splash.update()
    print( strftime('%H:%M:%S pythoncom') )
    import pythoncom
    splash.update()
    print( strftime('%H:%M:%S matplotlib') )
    import matplotlib.backends.backend_tkagg
    import matplotlib.animation
    splash.update()
    print( strftime('%H:%M:%S mpl_toolkits.axes_grid1') )
    import mpl_toolkits.axes_grid1
    splash.update()
    print( strftime('%H:%M:%S mpl_toolkits.mplot3d') )
    import mpl_toolkits.mplot3d
    splash.update()
    print( strftime('%H:%M:%S idlelib') )
    if sys.version_info > (3,6):
        import idlelib.tooltip
    else:
        import idlelib.ToolTip
    splash.update()
    print( strftime('%H:%M:%S pyperclip') )
    import pyperclip
    splash.update()
    print( strftime('%H:%M:%S wmi') )
    import wmi
    splash.update()
    try:
        print( strftime('%H:%M:%S numba') )
        import numba
        numba_is_available = True
        splash.update()
    except:
        print( strftime('%H:%M:%S failed to import numba') )

    splash.update()
    root.withdraw()

    print( strftime('%H:%M:%S GuiMain') )
    import molass_legacy.SerialAnalyzer
    import molass_legacy.AutorgKek
    from molass_legacy.KekLib.ArgParser import ArgParser
    from molass_legacy._MOLASS.GuiMain import GuiMain

    ap = ArgParser(sys.argv)
    kwargs = ap.get_dict()
    guimain = GuiMain(root, **kwargs)

    root.mainloop()

if __name__ == '__main__':
    # main must not be called in multiprocessing
    import sys
    import os
    this_dir = os.path.dirname(os.path.abspath(__file__))
    legacy_dir = os.path.join(this_dir, "..")
    molass_dir = os.path.join(legacy_dir, "..", "molass-library")
    sys.path.insert(0, legacy_dir)
    sys.path.insert(0, molass_dir)
    from molass_legacy import get_version
    print(f'MOLASS Legacy {get_version()}')
    from molass import get_version
    print(f'MOLASS Library {get_version()}')
    molass_gui()



# coding: utf-8
"""

    LinRegSmoother.py

    Copyright (c) 2017, Masatsuyo Takahashi, KEK-PF
"""
import numpy        as np
from scipy          import stats

REGRESS_SIZE    = 9
REGHALF_SIZE    = REGRESS_SIZE // 2

class Smoother:
    def __init__( self, x, y ):
        sy_list = []
        start   = 0
        stop    = len(x) - REGRESS_SIZE
        for i in range( start, stop ):
            slice_ = slice( i, i + REGRESS_SIZE )
            x_ = x[slice_]
            y_ = y[slice_]
            slope, intercept, r_value, p_value, std_err = stats.linregress( x_, y_ )
            if i == start:
                for j in range( 0, REGHALF_SIZE ):
                    sy = slope * x_[j] + intercept
                    sy_list.append( sy )

            sy = slope * x_[ REGHALF_SIZE ] + intercept
            sy_list.append( sy )

            if i == stop-1:
                for j in range( REGHALF_SIZE, REGRESS_SIZE ):
                    sy = slope * x_[j] + intercept
                    sy_list.append( sy )

        self.sy = np.array( sy_list )

    def __call__( self, x ):
        return self.predict( x )

    def predict( self, x ):
        return self.sy



# coding: utf-8
"""

    ModeledCurve.py

        evaluation of simularity between curves

    Copyright (c) 2021, SAXS Team, KEK-PF

"""
import numpy as np
from scipy.optimize import minimize
from matplotlib.patches import Rectangle
import molass_legacy.KekLib.DebugPlot as plt
from .RobustCurve import RobustCurve, plot_curve
from .ElutionModels import egh, emg

class Model:
    def __init__(self, name, func, params):
        self.name = name
        self.func = func
        self.params = params

    def __call__(self, x):
        return self.func(x, *self.params)

class FitHistory:
    def __init__(self, x, slice_history, model_history, params_history):
        self.x = x
        self.slice_history = slice_history
        self.model_history = model_history
        self.params_history = params_history

    def get_params_history(self):
        return self.params_history

    def get_shot(self, n):
        return self.slice_history[n], self.model_history[n]

    def get_peaktop_index(self):
        slice_ = self.slice_history[0]
        pti = (slice_.start + slice_.stop)//2
        return pti

    def plot_params_history(self, ax):
        params_array = np.array(self.params_history)
        pti = self.get_peaktop_index()
        num_iter = params_array.shape[0]
        for j in range(4):
            ax.plot(self.x[pti:pti+num_iter], params_array[:,j]/params_array[0,j], ':')

class ModeledCurve(RobustCurve):
    def __init__(self, x, y, depth=2, debug=False):
        RobustCurve.__init__(self, x, y, debug=debug)
        self.depth = depth
        self.children = []
        self.pgs = self.get_peak_models()

    def fit_peak(self, pno, func, allow=0.005):
        return self.fit_locally(self.pgs[pno], func, allow)

    def fit_locally(self, pg, func, allow):
        f, t = self.get_next_range(pg.pt_slice.start, pg.pt_slice.stop)
        max_iter = len(self.y)//2
        print("pg.params=", pg.params)
        h, m, s = pg.params
        pti = (f+t)//2
        params = (self.y[pti], m, s, 0)
        bounds = ((0, None), (m-s, m+s), (0, None), (-5*s, 5*s))
        slice_hishoty = []
        model_history = []
        params_history = []
        for k in range(max_iter):
            nf, nt = self.get_next_range(f, t)
            print([k], (f,t), (nf, nt))
            if nf == f and nt == t:
                break
            slice_ = slice(f, t+1)
            x_ = self.x[slice_]
            y_ = self.y[slice_]

            def objective(p):
                return np.sum((func(x_, *p) - y_)**2)

            result = minimize(objective, params, bounds=bounds)
            slice_hishoty.append(slice_)
            model_history.append(Model('egh', egh, result.x))
            params_history.append(result.x)
            max_deviation = np.max(np.abs(result.x/params_history[0] - 1))
            print([k], max_deviation)
            if max_deviation > allow:
                break
            params = result.x
            f, t = nf, nt
        return FitHistory(self.x, slice_hishoty, model_history, params_history)

    def get_side_residual(self, side, history):
        pti = history.get_peaktop_index()
        slice_ = slice(0, pti) if side == 0 else slice(pti, len(self.x))
        x_ = self.x[slice_]
        _, model = history.get_shot(0)
        y_ = self.y[slice_] - model(x_)
        return ModeledCurve(x_, y_, debug=False)

    def get_next_range(self, f, t):
        nf = max(0, f-1)
        nt = min(len(self.y)-1, t+1)
        while self.y[nf] > self.y[nt]:
            if nf > 0:
                nf -= 1
            else:
                break
        while self.y[nf] < self.y[nt]:
            if nt < len(self.y)-1:
                nt += 1
            else:
                break
        return nf, nt

    def plot_shot(self, ax, shot):
        x = self.x
        y = self.y
        ax.plot(x, y)
        slice_, model = shot
        my = model(x)
        ax.plot(x, my)
        ymin, ymax = ax.get_ylim()
        ax.set_ylim(ymin, ymax)
        f = x[slice_.start]
        t = x[slice_.stop-1]
        p = Rectangle(
                (f, ymin),      # (x,y)
                t - f,          # width
                ymax - ymin,    # height
                facecolor   = 'cyan',
                alpha       = 0.2,
            )
        ax.add_patch(p)

        ry = y - my
        ax.plot(x, ry)

    def plot(self, ax):
        plot_curve(ax, self)



"""
    UvCorrectorIllust.py

    Copyright (c) 2021-2022, SAXS Team, KEK-PF
"""
import numpy as np
import molass_legacy.KekLib.DebugPlot as plt
from molass_legacy.Peaks.EghSupples import egh, d_egh

def illust():
    with plt.Dp():
        x = np.arange(600)
        sigma = 20
        params = np.array([1, 300, sigma, 30])
        fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(12, 5))
        fig.suptitle("UV Distortion Model", fontsize=20)
        ax1.set_title("egh", fontsize=16)
        ax2.set_title("reflected derivative of egh", fontsize=16)

        for tau in [20, 30, 40]:
            params[2] = tau
            ax1.plot(x, egh(x, *params), label="sigma=%g, tau=%g" % (sigma, tau))
            params_ = params.copy()
            params_[0] = -1
            ax2.plot(x, d_egh(x, *params_), label="sigma=%g, tau=%g" % (sigma, tau))
        for ax in [ax1, ax2]:
            ax.legend()
        fig.tight_layout()
        plt.show()

def debug_plot(title_words, uv_corrector, x, y, params, fit_slice, init_params=None, fig_file=None):
    from DataUtils import get_in_folder
    from molass_legacy.Elution.CurveUtils import simple_plot
    print("params=", params)
    in_folder = get_in_folder()
    title = "UV Correction with %s for %s" % (title_words, in_folder)

    slope, intercept = params[0:2]
    dy_ = uv_corrector.correction_curve(x, *params)
    y_ = uv_corrector.get_corrected_y(params=params)
    if False:
        egh_params = params[2:].copy()
        if egh_params[0] < 0:
           egh_params[0] *= -1

    with plt.Dp():
        fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(12,5))
        fig.suptitle(title, fontsize=20)
        ax1.set_title("Elution at λ=%g" % 280, fontsize=16)
        ax2.set_title("Elution at λ=%g" % 340, fontsize=16)

        simple_plot(ax1, uv_corrector.curve1, legend=False)
        # ax1.plot(x, egh(x, *egh_params), label="egh")
        ax1.legend()

        if init_params is not None:
            scale = params[2]/init_params[2]
            tx = np.average(ax1.get_xlim())
            ty = np.average(ax1.get_ylim())
            ax1.text(tx, ty, "Scale: %.3g" % scale, ha="center", fontsize=30, alpha=0.3)

        ax2.plot(x, y, alpha=0.5, label="data")
        ax2.plot(x, dy_, color="cyan", lw=2, label="distotion model")
        ax2.plot(x, y_, color="green", alpha=0.5, label="corrected data")
        ax2.plot(x, x*slope + intercept, ":", color="red", lw=2, label="linear regression")

        ymin, ymax = ax2.get_ylim()
        ax2.set_ylim(ymin, ymax)
        for i, p in enumerate([fit_slice.start, fit_slice.stop]):
            label = "fit range ends" if i == 0 else None
            ax2.plot([p, p], [ymin, ymax], ":", color="gray", label=label)

        ax2.legend()
        fig.tight_layout()
        if fig_file is None:
            plt.show()
        else:
            from time import sleep
            plt.show(block=False)
            fig = plt.gcf()
            fig.savefig(fig_file)
            sleep(1)



# coding: utf-8
"""
    ConjugateGradientDemo.py

    Copyright (c) 2019-2020, SAXS Team, KEK-PF
"""
from bisect import bisect_right
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.gridspec import GridSpec
import mpl_toolkits.mplot3d.axes3d as p3
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
import matplotlib.patches as mpl_patches
from matplotlib.widgets import Button

import logging
from molass_legacy.KekLib.OurTkinter import Tk, Dialog
from DataUtils import cut_upper_folders
from DecompUtils import CorrectedBaseline, make_range_info_impl, decompose_elution_better, get_peak2elem_dict
from molass_legacy.Models.ElutionCurveModels import EMGA
from molass_legacy.ElutionDecomposer import ElutionDecomposer
from ExtrapolationSolver import ExtrapolationSolver
from molass_legacy.PeaksetSelector import PeakSetSelector
from molass_legacy.DataStructure.AnalysisRangeInfo import convert_to_paired_ranges
from molass_legacy.KekLib.OurMatplotlib import get_color
import molass_legacy.KekLib.DebugPlot as dplt
from molass_legacy._MOLASS.SerialSettings import get_setting, set_setting
from ConjugateGradientThreeD import GdTreedPlot

ENABLE_GD_TRACE = True
TITLE_FONTSIZE  = 16
LINEWIDTH       = 3
DEMO_MIN_ANGLE  = 0.03
DEMO_ANGLE_SIZE = 40        # points

default_colors = plt.rcParams['axes.prop_cycle'].by_key()['color']

class ConjugateGradientDemo(Dialog):
    def __init__(self, parent, in_folder, sd, mapper):
        self.logger = logging.getLogger(__name__)
        self.in_folder = cut_upper_folders(in_folder)
        self.sd = sd
        self.mapper = mapper
        self.popup_menu = None
        self.make_demo_range_info()
        self.use_elution_models_save = get_setting('use_elution_models')
        set_setting('use_elution_models', 1)

        Dialog.__init__( self, parent, "Gradient Descent Demo", visible=False )

    def __del__(self):
        # print('__del__: restore use_elution_models')
        set_setting('use_elution_models', self.use_elution_models_save)

    def make_demo_range_info(self):
        self.corbase_info = CorrectedBaseline(self.sd, self.mapper)
        self.decomp_ret = ret = decompose_elution_better(self.corbase_info, self.mapper, EMGA(),
                                                    logger=self.logger)

        self.x = ret.x
        self.y = ret.y
        self.uv_y = ret.uv_y
        self.opt_recs    = ret.opt_recs
        self.opt_recs_uv = ret.opt_recs_uv
        self.num_eltns = len(self.opt_recs_uv)

        ret_info = ret.get_range_edit_info(logger=self.logger)
        self.range_info = make_range_info_impl(ret.opt_recs_uv, ret_info)

        ret = convert_to_paired_ranges(self.range_info)
        self.cnv_ranges = ret[0]
        selector = PeakSetSelector(self.cnv_ranges, self.mapper.x_curve)
        self.peakset_info, elem_no = selector.select_demo_ranges_for_gd(self.uv_y, self.opt_recs_uv)
        opt_rec_uv = self.opt_recs_uv[elem_no]
        self.evaluator = opt_rec_uv[1]
        self.elm_color = get_color(len(self.opt_recs_uv) - 1)
        self.eval_range = self.evaluator.get_range_params(self.x)

        self.q = self.sd.intensity_array[0,:,0]
        self.qf = bisect_right(self.q, DEMO_MIN_ANGLE)
        self.qsize = DEMO_ANGLE_SIZE
        self.qt = self.qf + self.qsize

    def show(self):
        self._show()

    def body(self, body_frame):
        cframe = Tk.Frame(body_frame)
        cframe.pack()

        self.fig = fig = plt.figure(figsize=(21, 7))

        gs = GridSpec( 2, 3 )

        self.mpl_canvas = FigureCanvasTkAgg( fig, cframe )
        self.mpl_canvas_widget = self.mpl_canvas.get_tk_widget()
        self.mpl_canvas_widget.pack( fill=Tk.BOTH, expand=1 )
        self.mpl_canvas.draw()

        self.ax01 = self.fig.add_subplot(gs[0, 0])
        self.ax02 = self.fig.add_subplot(gs[1, 0])
        self.ax1 = self.fig.add_subplot(gs[:, 1], projection='3d')
        self.ax2 = self.fig.add_subplot(gs[:, 2], projection='3d')
        fig.suptitle("Gradient Descent Demo using " + self.in_folder, fontsize=20)

        self.fig.subplots_adjust( top=0.9, bottom=0.08, left=0.05, right=0.97, wspace=0.1, hspace=0.3 )
        self.draw_elution()
        self.solve_extrapolation()
        self.draw_scattering()
        self.draw_data_3d()
        self.draw_gd_3d()
        pos = plt.axes([0.93, 0.04, 0.04, 0.04])
        self.animate_btn = Button(pos, 'Animate')
        self.animate_btn.on_clicked(self.show_animation)

    def buttonbox( self ):
        box = Tk.Frame(self)
        box.pack()
        w = Tk.Button(box, text="OK", width=10, command=self.ok, default=Tk.ACTIVE)
        w.pack(side=Tk.LEFT, padx=5, pady=5)
        self.bind("<Return>", self.ok)

    def draw_elution(self):
        ax = self.ax01
        ax.set_title("Focused range in the UV-elution curve", fontsize=TITLE_FONTSIZE)

        x = self.x
        y = self.uv_y
        ax.plot(x, y, color='blue')
        ax.plot(x, self.evaluator(x), ':', color=self.elm_color, linewidth=LINEWIDTH)

        ymin, ymax = ax.get_ylim()
        f, _, t = self.eval_range
        p = mpl_patches.Rectangle(
                (f, ymin),  # (x,y)
                t - f,   # width
                ymax - ymin,    # height
                facecolor   = 'cyan',
                alpha       = 0.2,
            )
        ax.add_patch( p )

    def solve_extrapolation(self):
        from PreviewData import PreviewData, PreviewOptions
        self.x_curve = self.mapper.x_curve
        self.data = self.sd.intensity_array[:,:,1].T
        self.error = self.sd.intensity_array[:,:,2].T
        conc_factor = 5

        pdata = PreviewData(sd=self.sd, mapper=self.mapper, paired_ranges=self.cnv_ranges)
        popts = PreviewOptions(use_elution_models=False)

        self.solver = ExtrapolationSolver(pdata, popts)
        start, _, stop = self.eval_range

        # TODO: "de-couple" cnv_ranges, peakset_info from the solver

        demo_info = np.array([self.qf, self.qt]) if ENABLE_GD_TRACE else None
        self.solver_result = self.solver.extrapolate_wiser( start, stop, self.peakset_info,
                                demo_info=demo_info,
                                )

    def draw_scattering(self):
        A, B, Z, E, _, C = self.solver_result

        if False:
            f, _, t = self.eval_range
            fig = dplt.figure()
            axd = fig.gca()
            axd.set_title("solver_result C[0,:]")
            j = np.arange(f, t)
            axd.plot(j, C[0,:])
            fig.tight_layout()
            dplt.show()

        ax = self.ax02
        ax.set_title("Focused range in the Xray-scattering curve", fontsize=TITLE_FONTSIZE)

        _, top_x, _ = self.eval_range
        y = self.data[:,top_x]
        ax.plot(self.q, y)
        self.compute_p2m_scale(A, y)
        ax.plot(self.q, A*self.p2m_scale, color='orange')

        qf_ = self.q[self.qf]
        qt_ = self.q[self.qt]
        self.p1 = zf_ = A[self.qf]*self.p2m_scale
        self.p2 = zt_ = A[self.qt]*self.p2m_scale

        qf_ = self.q[self.qf]
        qt_ = self.q[self.qt]
        self.ax02.plot([qf_, qf_], [0, zf_], color='red', alpha=0.3, linewidth=5)
        self.ax02.plot([qt_, qt_], [0, zt_], color='green', alpha=0.3, linewidth=5)

        ymin, ymax = ax.get_ylim()
        p = mpl_patches.Rectangle(
                (qf_, ymin),  # (x,y)
                qt_ - qf_,   # width
                ymax - ymin,    # height
                facecolor   = 'orange',
                alpha       = 0.2,
            )
        ax.add_patch( p )

    def compute_p2m_scale(self, A, y):
        k = bisect_right(self.q, 0.05)
        slice_ = slice(k-2, k+3)
        self.p2m_scale = np.average(y[slice_]) / np.average(A[slice_])
        print('p2m_scale=', self.p2m_scale)

    def draw_data_3d(self):
        A = self.solver_result[0]

        f, _, t = self.eval_range
        esize = t - f
        y = np.arange(f, t)

        elution_z = self.evaluator(y)
        self.elution_z = elution_z = elution_z/np.max(elution_z) * self.p2m_scale

        color = get_color(0)
        qf = self.qf
        qt = self.qt
        qsize = self.qsize

        qt1 = qt + 1
        q_ = self.q[qf:qt1]
        xx, yy = np.meshgrid(q_, y)

        ax = self.ax1
        ax.set_title("Differences between the data and the factorized surface", fontsize=TITLE_FONTSIZE)

        _, top_x, _ = self.eval_range
        zf_ = self.p1
        zt_ = self.p2
        qf_ = self.q[qf]
        qt_ = self.q[qt]

        ax.plot([qf_, qf_], [top_x, top_x], [0, zf_], color='red', alpha=0.3, linewidth=5)
        ax.plot([qf_], [top_x], [zf_], 'o', color='red', markersize=10)

        zz_ = []
        for i in range(qf, qt1):
            q_ = self.q[i]
            x = np.ones(esize) * q_
            z = self.data[i, f:t]
            ax.plot(x, y, z, 'o', color=color, markersize=1)
            z_ = elution_z * A[i]
            zz_.append( z_ )
            if i in [qf, qt]:
                self.draw_diff_lines(ax, q_, y, z, z_)

            if False:
                print('A[%d]=' % i, A[i])
                fig = dplt.figure()
                axd = fig.gca()
                axd.set_title( 'A[%d]=%.3g' % ( i, A[i] ) )
                axd.plot(y, z)
                axd.plot(y, z_)
                fig.tight_layout()
                dplt.show()

        zz = np.array(zz_).T
        ax.plot_surface(xx, yy, zz, color='orange', alpha=0.3)

        ax.plot([qt_, qt_], [top_x, top_x], [0, zt_], color='green', alpha=0.3,linewidth=5)
        ax.plot([qt_], [top_x], [zt_], 'o', color='green', markersize=10)

    def draw_diff_lines(self, ax, q_, y, z, z_):
        for k, j in enumerate(y):
            ax.plot([q_, q_], [j, j], [z[k], z_[k]], color='pink', linewidth=3)

    def draw_gd_3d(self):
        self.gd3d = GdTreedPlot(self, ENABLE_GD_TRACE)
        self.gd3d.draw(self.ax2, TITLE_FONTSIZE)

    def show_animation(self, event):
        print('show_animation')
        from ConjugateGradientThreeD import GdAnimationDialog
        dialog = GdAnimationDialog(self, self.gd3d)
        dialog.show()



"""
    Estimators.SdmEstimatorImpl.py

    Copyright (c) 2024-2025, SAXS Team, KEK-PF
"""
import numpy as np
import molass_legacy.KekLib.DebugPlot as plt
from molass_legacy._MOLASS.SerialSettings import get_setting
from molass_legacy.Models.Stochastic.RgReliability import determine_unreliables

def guess_exec_spec(peak_rgs, props, qualities):
    """
    for these settings, see also SecTheory.ColumnTypes.py, Optimizer.ptimizerSettings.py
    """
    unreliable_indeces = determine_unreliables(peak_rgs, qualities, props, trust_max_num=None)
    poresize_bounds = get_setting("poresize_bounds")
    init_N0 = get_setting("num_plates_pc")
    return {"unreliable_indeces": unreliable_indeces, 'poresize_bounds': poresize_bounds, 'init_N0': init_N0}

def edit_to_full_sdmparams(editor, sdm_params, corrected_rgs, uv_curve, debug=False):
    if debug:
        from importlib import reload
        import molass_legacy.Models.Stochastic.MonoporeUvScaler
        reload(molass_legacy.Models.Stochastic.MonoporeUvScaler)
        
    from molass_legacy.Models.Stochastic.DispersiveUtils import NUM_SDMCUV_PARAMS
    from molass_legacy.Models.Stochastic.DispersiveUvScaler import adjust_to_uv_scales

    # (xr_curve, D), rg_curve, (uv_curve, U) = editor.dsets

    N, K, x0, poresize, N0, tI = sdm_params[0:NUM_SDMCUV_PARAMS]
    T = K/N
    me = 1.5
    mp = 1.5
    xr_w = sdm_params[NUM_SDMCUV_PARAMS:]
    rgs = corrected_rgs
    sdmcol_params = np.array([N, K, x0, poresize, N0, tI])      # N, K, x0, poresize, N0, tI

    _, _, xr_x, xr_y, baselines = editor.get_curve_xy(return_baselines=True)
    a, b = editor.peak_params_set[-2:]
    uv_x = xr_x*a + b
    uv_y = uv_curve.spline(uv_x)

    uv_baseline = editor.get_uv_baseline_deprecated(xy=(uv_x, uv_y))    # 

    uv_y_ = uv_y - uv_baseline
    xr_y_ = xr_y - baselines[1]

    ret = adjust_to_uv_scales(xr_x, xr_y_, uv_x, uv_y_, sdm_params, corrected_rgs, debug=debug)
    if ret is None:
        return
    else:
        uv_w, uv_ty = ret

    uv_base_params = editor.get_uv_base_params(xyt=(uv_x, uv_y, uv_ty))
    param_list = [xr_w, editor.baseline_params[1], rgs, (a, b), uv_w, uv_base_params, xr_x[[0,-1]], sdmcol_params]
    init_params = np.concatenate(param_list)

    if debug:
        # self.logger.info("uv_base_params=%s, average(uv_ty)=%g", str(uv_base_params), np.average(uv_ty))
        uv_bl_computed = editor.base_curve_info[0](uv_x, uv_base_params, uv_ty)
        with plt.Dp():
            fig,(ax1, ax2) = plt.subplots(ncols=2, figsize=(12,5))
            fig.suptitle("compute_sdm_init_params")
            ax1.plot(uv_x, uv_y)
            ax1.plot(uv_x, uv_ty + uv_baseline, ":")
            ax1.plot(uv_x, uv_baseline, color="red")
            ax1.plot(uv_x, uv_bl_computed, color="yellow")
            ax2.plot(xr_x, xr_y)
            ax2.plot(xr_x, baselines[1], color="red")
            fig.tight_layout()
            plt.show()

    return init_params



# coding: utf-8
"""
    AbsorbanceViewerPub.py

    Copyright (c) 2018, Masatsuyo Takahashi, KEK-PF
"""
import numpy                as np
from bisect                 import bisect_right
import matplotlib.pyplot    as plt
import matplotlib.gridspec  as gridspec
from mpl_toolkits.mplot3d   import Axes3D
from matplotlib.backends.backend_tkagg  import FigureCanvasTkAgg
from molass_legacy.KekLib.OurTkinter             import Tk, Dialog
from molass_legacy.KekLib.TkUtils                import is_low_resolution
from molass_legacy.KekLib.BasicUtils             import get_caller_module
from molass_legacy.KekLib.TkSupplements          import tk_set_icon_portable
from molass_legacy.KekLib.OurMatplotlib          import NavigationToolbar
from molass_legacy._MOLASS.SerialSettings         import get_setting

DEBUG = False

class AbsorbanceViewerPub( Dialog ):
    def __init__( self, absorbance ):
        self.grab = 'local'     # used in grab_set
        self.absorbance     = absorbance
        self.wl_vector      = absorbance.wl_vector
        self.wvlen_lower    = 270
        self.wvlen_upper    = 350
        f = bisect_right( absorbance.wl_vector, self.wvlen_lower )
        t = bisect_right( absorbance.wl_vector, self.wvlen_upper )
        self.i_slice        = slice( f, t )
        self.applied    = None
        self.caller_module = get_caller_module( level=2 )

    def show( self, parent, button_labels=[ "OK", "Cancel" ], toolbar=False ):
        assert( len(button_labels) == 2 )
        self.button_labels  = button_labels
        self.toolbar    = toolbar

        Dialog.__init__( self, parent, "AbsorbanceViewer" )
        # TODO: self.resizable(width=False, height=False)

    def body( self, body_frame ):   # overrides parent class method
        tk_set_icon_portable( self, module=self.caller_module )

        cframe = Tk.Frame( body_frame )
        cframe.pack()

        figsize = ( 9, 8 ) if is_low_resolution() else ( 11, 10 )
        self.fig = fig = plt.figure( figsize=figsize )

        self.mpl_canvas = FigureCanvasTkAgg( fig, cframe )
        self.mpl_canvas.draw()
        self.mpl_canvas_widget = self.mpl_canvas.get_tk_widget()
        self.mpl_canvas_widget.pack( fill=Tk.BOTH, expand=1 )

        # it seems that draw_func should be called after the creation of mpl_canvas
        # in order to enable 3d-rotation of the figure
        ax1 = fig.add_subplot( 111, projection='3d' )
        # ax2 = fig.add_subplot( 122 )

        self.axes   = [ax1, None]

        self.draw()
        self.parent.config( cursor='' )

        if self.toolbar:
            self.toolbar = NavigationToolbar( self.mpl_canvas, cframe )
            self.toolbar.update()

        self.protocol( "WM_DELETE_WINDOW", self.ok )

    def buttonbox(self):
        '''add standard button box.

        override if you do not want the standard buttons
        '''

        box = Tk.Frame(self)

        w = Tk.Button(box, text=self.button_labels[0], width=10, command=self.ok, default=Tk.ACTIVE)
        w.pack(side=Tk.LEFT, padx=5, pady=5)
        self.ok_button = w
        w = Tk.Button(box, text=self.button_labels[1], width=10, command=self.cancel)
        w.pack(side=Tk.LEFT, padx=5, pady=5)
        self.cancel_button = w

        self.bind("<Return>", self.ok)
        self.bind("<Escape>", self.cancel)

        box.pack()

    def ok( self, *args ):
        self.applied  = True
        self.destroy()

    def cancel( self, *args ):
        self.applied  = False
        self.destroy()

    def draw( self ):
        ax1 = self.axes[0]

        self.draw_3d( ax1 )
        # self.draw_3d_detail( ax1 )
        self.draw_annotion( ax1 )

        self.fig.tight_layout()

    def draw_3d( self, ax, zlim=None ):
        lower   = self.wvlen_lower
        upper   = self.wvlen_upper

        ax.set_title( "" )
        ax.set_xlabel( '\nwave length (nm)' )
        ax.set_ylabel( '\nsequential №' )
        ax.set_zlabel( '\nabsorbance' )
        ax.set_xlim( self.wvlen_lower - 10 , self.wvlen_upper + 10 )

        if zlim is None:
            # zlim = ( -0.05, 0.2 )
            # ax.set_zlim( zlim )
            pass
        else:
            ax.set_zlim( zlim )

        data = self.absorbance.data
        print( 'data.shape=', data.shape )

        size = data.shape[1]

        peak_j = None
        tail_i = self.absorbance.tail_index
        absorbance_baseline_type = get_setting( 'absorbance_baseline_type' )
        for i, wv in enumerate( self.absorbance.wl_vector ):
            if i == tail_i:
                break
            else:
                if wv < self.wvlen_lower or wv > self.wvlen_upper:
                    continue
                if i != self.absorbance.index and i % 2 > 0:
                    continue

            X = np.ones( size ) * wv
            Y = np.arange( size )
            Z = data[i,:]
            if i == self.absorbance.index:
                alpha   = 1
                color   = 'blue'
            else:
                alpha   = 0.2
                color   = '#1f77b4'
            ax.plot( X, Y, Z, color=color, alpha=alpha )

            if absorbance_baseline_type == 0 or self.absorbance.base_curve is None:
                continue

            if i  == tail_i:
                slice_ = slice( self.absorbance.rail_points[0], self.absorbance.rail_points[1]+1 )
                Y_ = Y[slice_]
                curve_size = len(self.absorbance.tail_curve)
                for ii in range( i-5, i+6, 2 ):
                    X_ = np.ones( curve_size ) * self.absorbance.wl_vector[ii]
                    ax.plot( X_, Y_, self.absorbance.tail_curve, color='red', alpha=0.5 )

                X_ = np.ones( curve_size ) * self.absorbance.std_wvlen
                ax.plot( X_, Y_, self.absorbance.base_curve, color='red', alpha=0.5 )

    def draw_3d_detail( self, ax ):
        wvlen   = self.absorbance.wl_vector
        lower   = self.wvlen_lower
        upper   = self.wvlen_upper
        absorbance  = self.absorbance

        wvlen_ = wvlen[self.i_slice]

        if False:
            lower_j  = absorbance.a_curve.lower_abs
            middle_j = absorbance.a_curve.middle_abs
            upper_j  = absorbance.a_curve.upper_abs
        else:
            n       = absorbance.a_curve.primary_peak_no
            info    = absorbance.a_curve.peak_info[n]
            lower_j  = info[0]
            middle_j = int(info[1] + 0.5)
            upper_j  = info[2]

        section_array = []
        colors = [ 'green', 'orange', 'green', 'purple', 'purple' ]
        for i, j in enumerate( [ lower_j, middle_j, upper_j, absorbance.jump_j, absorbance.right_jump_j ] ):
            print( 'draw_3d_absorbance_detail: (i, j)=', (i, j) )
            if j is None:
                continue

            X = wvlen_
            Y = np.ones( len(wvlen_) ) * j
            Z = self.absorbance.data[self.i_slice,j]
            section_array.append( [ j, Z ] )
            color = colors[i]
            ax.plot( X, Y, Z, ':', color=color )

        # self.draw_wave_length_side_2d( ax_2d,  wvlen_, section_array )

    def draw_annotion( self, ax ):
        wvlen   = self.absorbance.wl_vector
        absorbance  = self.absorbance

        bline = absorbance.get_bottomline_matrix()

        xmin, xmax = ax.get_xlim()
        zmin, zmax = ax.get_zlim()
        xoffset = ( xmax - xmin ) * 0.01
        zoffset = ( zmax - zmin ) * 0.005

        Y = np.arange( absorbance.data.shape[1] )
        for i in [0, 3]:
            wvlen = absorbance.reg_wvlens[i]
            X = np.ones( absorbance.data.shape[1] ) * wvlen
            Z = bline[:, i]
            if i == 3:
                color   = 'red'     # 'cyan'
                alpha   = 1
            else:
                color   = 'red'
                alpha   = 1
            ax.plot( X, Y, Z, ':', color=color, alpha=alpha )
            if i in [0,3]:
                ax.text( wvlen - xoffset, Y[-1], Z[-1] + zoffset, 'λ=%g' % wvlen )

        A, B, C = absorbance.baseplane_params

        xmin, xmax = absorbance.reg_wvlens[[0, 3]]
        ymin, ymax = 0, absorbance.data.shape[1]-1

        """
        xx, yy = np.meshgrid( np.linspace(xmin, xmax, 10), np.linspace(ymin, ymax, 10) )
        zz = xx * A + yy * B + C
        ax.plot_surface(xx, yy, zz, alpha=0.3, color='red' )
        """
        yy = np.array( [ymin, ymax] )

        for i, wv in enumerate( self.absorbance.wl_vector ):
            if wv < xmin:
                continue
            if wv > xmax:
                break

            if i % 2 > 0:
                continue

            xx = np.array( [ wv, wv ] )
            zz = xx * A + yy * B + C

            ax.plot( xx, yy, zz, color='red', alpha=0.2 )



"""
    SpotDebugger.py

    Copyright (c) 2023, SAXS Team, KEK-PF
"""
from molass_legacy.KekLib.OurTkinter import Tk, Dialog

parents = []

class SpotDebugger(Dialog):
    def __init__(self, parent, debug_info):
        if parent is None:
            from molass_legacy.KekLib.TkUtils import adjusted_geometry
            parent = Tk.Tk()
            parents.append(parent)
            parent.geometry( adjusted_geometry( parent.geometry() ) )
            parent.withdraw()
            parent.update()

            """
            FIXME:
                _tkinter.TclError: grab failed: another application has grab occurred in ElutionDecomposer.add_other_peaks
            """

        self.debug_info = debug_info
        Dialog.__init__(self, parent, "Spot Debugger", visible=False)

    def show(self):
        self._show()

    def body(self, body_frame):
        button  = Tk.Button(body_frame, text="Debug", command=self.run_debug)
        button.pack()

    def run_debug(self):
        from importlib import reload
        import Tools.DebugCushion
        reload(Tools.DebugCushion)
        from molass_legacy.Tools.DebugCushion import debug_impl
        debug_impl(self.debug_info)



"""
    DebugUtils.py

    Copyright (c) 2022, SAXS Team, KEK-PF
"""
import os
from termcolor import colored

os.system('color')  # from https://stackoverflow.com/questions/287871/how-do-i-print-colored-text-to-the-terminal

last_lines = None

def show_call_stack(line_prefix, stop_level=None, indented_lines=6, indented_only=False):
    """
    for usage, see Trimming.AutoRestrictor.py where this is used for the first time
    """
    import inspect
    global last_lines

    stack = inspect.stack()[1:stop_level]

    if not indented_only:
        for frm in stack:
            print("%s%s %s (%d)" % (line_prefix, frm.filename, frm.function, frm.lineno))

    n = indented_lines
    stored_lines = []
    for k in range(n):
        i = n - k - 1
        frm = stack[i]
        line = "%s%s (%d)" % (" "*(4*k), frm.function, frm.lineno)
        stored_lines.append(line)
        color_print = False
        if last_lines is not None:
            if k < len(last_lines):
                if line != last_lines[k]:
                    color_print = True

        if color_print:
            print(colored(line, 'yellow'))
        else:
            print(line)

    last_lines = stored_lines



"""
    Trimming.AutoRestrictorProxy.py

    Copyright (c) 2024-2025, SAXS Team, KEK-PF
"""
from  molass_legacy.Trimming.AutoRestrictor import get_exact_emg_peaks_list, SIGMA_POINT_RATIO

class AutoRestrictorProxy:
    def __init__(self, sd, uv_restrict_list, xr_restrict_list):
        # self.result_info = sd, curves, emg_peaks_list, old_e_restricts, ret_info, SIGMA_POINT_RATIO, self.mpeaks
        # minimum info for trimming_result_plot_impl
        uv_curve = sd.get_uv_curve()
        xr_curve = sd.get_xr_curve()
        curves = uv_curve, xr_curve

        uv_peaks, xr_peaks = get_exact_emg_peaks_list(uv_curve, xr_curve, debug=False)
        emg_peaks_list = uv_peaks, xr_peaks

        old_e_restricts = None,None
        ret_info = [uv_restrict_list, xr_restrict_list]
        mpeaks = None
        self.result_info = sd, curves, emg_peaks_list, old_e_restricts, ret_info, SIGMA_POINT_RATIO, mpeaks


"""

    ファイル名：   SphericalModel.py

    処理内容：

        ゼロ濃度外挿シミュレーションのプロット

    Copyright (c) 2017-2023, Masatsuyo Takahashi, KEK-PF

"""

import numpy            as np
from scipy.stats        import norm
from bisect             import bisect_right
from molass_legacy.KekLib.NumpyUtils         import np_savetxt

#   feigin-svergun-1987 (1.27)
def f( sR ):
     return 3 * ( np.sin( sR ) - sR * np.cos( sR ) ) / ( sR )**3

class XrayScattering:
    def __init__( self, N, R ):
        self.N  = N
        self.R  = R
        self.V  = 4 * np.pi * R**3 / 3
        self.v1 = self.V / N        # volume per particle

    def term1( self, x ):
        sR  = x * self.R
        return self.N * f(sR)**2

    def term2( self, x ):
        sR  = x * self.R
        return 8 * self.V / self.v1 * f( 2*sR )

    def __call__( self, x ):
        return self.term1( x ) * ( 1 - self.term2( x ) )

class UvAbsorbance:
    def __init__( self, num_elution_points, wl_vector, model_curve, scale=1 ):
        self.num_elution_points = num_elution_points
        self.wl_vector = wl_vector
        self.model_curve = model_curve
        self.scale = scale

    def generate_data( self ):
        x   = np.linspace(-7, 7, self.num_elution_points)
        y   = norm.pdf( x, 0, 1 )
        y_  = y / norm.pdf( 0, 0, 1 ) * self.scale
        data_list = [ self.wl_vector ]
        for i in range(len(x)):
            data_list.append( y_[i]*self.model_curve )

        self.data_array = np.array( data_list )

    def save( self, file ):
        fh = open( file, "wb" )
        line = "\t" + "\t".join( [ str(i) for i in range(self.num_elution_points+1) ] ) + '\n'
        fh.write( bytearray( line, 'cp932' ) )  # utf8 or cp932
        for j in range(self.data_array.shape[1]):
            line = "\t".join( [ '%g' % f for f in self.data_array[:,j] ] ) + '\n'
            fh.write( bytearray( line, 'cp932' ) )  # utf8 or cp932
        fh.close()

    def get_elutioncurve( self, lambda_, num_points ):
        j = bisect_right( self.wl_vector, lambda_ )
        indeces = np.linspace( 0, self.num_elution_points-1, num_points, dtype=int )
        return self.data_array[ 1:, j ][indeces]

class SerialXrayScattering:
    def __init__( self, R, q_vector, c_vector, c_scale=1, i_scale=1 ):
        self.R  = R
        self.q_vector   = q_vector
        self.c_vector   = c_vector
        self.c_scale    = c_scale
        self.i_scale    = i_scale

    def generate( self ):
        intensity_list = []
        error_list = []
        x = self.q_vector

        for c in self.c_vector:
            c_ = c * self.c_scale
            model = XrayScattering( c_, self.R )
            noise_level = 1/( 0.01 + c)
            intensity = model( x ) * ( 1 + ( noise_level * 1e-4 + ( x - 0.05 ) **2 )* np.random.normal( 0, noise_level, len(x) ) )
            e = np.ones( len(x) ) * 1e-5 * noise_level
            intensity_list.append( intensity )
            error_list.append( e )

        self.intensity_array = np.array( intensity_list ) * self.i_scale
        self.error_array = np.array( error_list ) * self.i_scale

    def save( self, folder, name='Sample_000.dat' ):
        for n in range(len(self.c_vector)):
            path = folder + '/' + name.replace( '000', '%03d' % (n) )

            fh = open( path, "wb" )
            fh.write( bytearray( '# q intensity error\n', 'cp932' ) )
            fh.close()

            data = np.vstack( [ self.q_vector, self.intensity_array[n,:], self.error_array[n,:] ] )
            np_savetxt( path, data.T, mode="a" )



"""
    ScatteringBasesurface.py

    Copyright (c) 2017-2023, Masatsuyo Takahashi, KEK-PF
"""
import numpy                as np
import logging
import OurStatsModels       as sm
from ScatteringBaseUtil     import SMALL_ANGLE_LIMIT
from molass_legacy.Baseline.ScatteringBaseline     import ScatteringBaseline
from ScatteringBasecurve    import ScatteringBasecurve
from ScatteringBasespline   import ScatteringBasespline
# from molass_legacy.KekLib.NumpyUtils             import np_savetxt
from molass_legacy._MOLASS.SerialSettings         import get_setting
from molass_legacy.SerialAnalyzer.ElutionBaseCurve       import ElutionBaseCurve
from molass_legacy.KekLib.ExceptionTracebacker   import ExceptionTracebacker

DEBUG_PLOT  = False

NUM_SLOPE_ESTIMATION    = 10
NUM_SAMPLE_CURVES       = 20
VERY_SMALL_ANGLE_LIMIT  = 0.03

def is_to_plot( self, i, q, very_small_angle_only=False ):
    # return q >= 0.01 and ( i == self.index or i % 8 == 0 )
    if very_small_angle_only:
        return q < VERY_SMALL_ANGLE_LIMIT
    else:
        return i == self.index or i % 8 == 0

class ScatteringBasesurface:
    def __init__( self, qvector, index, data,
                    inty_curve_y=None, parent=None, baseline_degree=None ):
        self.qvector    = qvector
        self.index      = index
        self.data       = data
        self.parent     = parent
        self.logger     = logging.getLogger( __name__ )

        self.size = size = self.data.shape[0]

        if baseline_degree is None:
            baseline_degree = get_setting( 'baseline_degree' )

        self.baseline_degree = baseline_degree

        # TODO: pass size_sigma instead of inty_curve_y
        #       See also apply_baseline_correction_impl
        if inty_curve_y is None:
            size_sigma = None
        else:
            ecurve = ElutionBaseCurve( inty_curve_y )
            size_sigma = ecurve.compute_size_sigma()
            print( 'default size_sigma=', size_sigma )

        self.size_sigma = size_sigma

        self.Y = np.linspace( 0, size-1, NUM_SAMPLE_CURVES, dtype=int )

        iv, line_array = self.compute_lines()

        self.iv = iv
        self.line_array = line_array

        self.iv_vsa = None
        self.line_array_vsa = None
        self.average_curve = None
        # self.compute_slope_param_fg()
        # self.compute_average_curve()

    def compute_lines( self, very_small_angle_only=False ):
        j = np.arange( self.data.shape[0] )
        i_list = []
        line_list = []
        continue_plot = True
        for i, q in enumerate( self.qvector ):
            if not is_to_plot( self, i, q, very_small_angle_only=very_small_angle_only ):
                continue

            i_list.append(i)
            Z = self.data[:,i,1]

            ecurve = ElutionBaseCurve( Z )

            if q < SMALL_ANGLE_LIMIT:
                p_final = ecurve.compute_base_percentile_offset( q, return_simple=True )
            else:
                p_final = ecurve.compute_base_percentile_offset( q, size_sigma=self.size_sigma )

            iterate = 2 if q < SMALL_ANGLE_LIMIT else 1
            baseline = np.zeros( len(Z) )
            try:
                for k in range(iterate):
                    if self.baseline_degree == 1:
                        sbl = ScatteringBaseline( Z - baseline )
                        A, B = sbl.solve( p_final=p_final )
                        z = A*j + B
                    elif self.baseline_degree == 2:
                        sbl = ScatteringBasecurve( Z - baseline )
                        F, D, E = sbl.solve( p_final=p_final )
                        z = F*j**2 + D*j + E
                    else:
                        sbl = ScatteringBasespline( Z - baseline, q=q )
                        sbl.solve( p_final=p_final )
                        z = sbl.get_baseline( j )

                    baseline += z
            except:
                etb = ExceptionTracebacker()
                self.logger.warning( str(etb) + ' with i=' + str(i) + ', k=' + str(k) + ', Z=' + str(Z) )
                # occured in Yonezawa-san data. See the 20170831 e-mail
                # self.logger.warning( etb )

            line_list.append( baseline[self.Y] )
            if DEBUG_PLOT and continue_plot:
                continue_plot = sbl.debug_plot( title='Debug at Q[%d] = %.3g' % (i, q), parent=self.parent )

        return np.array( i_list ), np.array( line_list )

    def compute_slope_param_fg( self ):
        # np_savetxt( 'line_array.csv', self.line_array )
        x = []
        y = []
        for i in self.iv[:NUM_SLOPE_ESTIMATION]:
            print( 'i=', i )
            for k, j in enumerate( self.Y ):
                x.append( j )
                y.append( self.line_array[i,k]/self.line_array[i,0] - 1 )

        if True:
            from DebugCanvas    import DebugCanvas
            def xy_plot( fig ):
                ax = fig.add_subplot( 111 )
                for i_ in range(NUM_SLOPE_ESTIMATION):
                    start = i_ * len( self.Y )
                    slice_ = slice( start, start+len( self.Y ) )
                    ax.plot( x[slice_], y[slice_], 'o', label=str(i_) )

                ax.legend()
                fig.tight_layout()

            dc = DebugCanvas( "Debug x, y", xy_plot, toolbar=True )
            dc.show()

        model = sm.OLS(y, x)
        results = model.fit()
        print( results.params )
        self.fg = results.params[0]
        return self.fg

    def compute_average_curve( self ):
        normalized_curves_list = []
        for k, j in enumerate( self.Y ):
           curve = self.line_array[ self.iv, k ] / ( self.fg * j + 1 )
           normalized_curves_list.append( curve )

        normalized_curves = np.array( normalized_curves_list )
        print( 'normalized_curves.shape=', normalized_curves.shape )
        self.average_curve = np.average( normalized_curves, axis=0 )
        print( 'self.average_curve.shape=', self.average_curve.shape )

    def plot( self, ax1, ax2=None, very_small_angle_only=False ):
        if ax2 is not None:
            ax2.set_title( "Background curves made with low percentile method" )

        if very_small_angle_only:
            iv = np.where( self.qvector < VERY_SMALL_ANGLE_LIMIT )[0]
            if self.line_array_vsa is None:
                iv, line_array = self.compute_lines( very_small_angle_only=True )
                self.iv_vsa = iv
                self.line_array_vsa = line_array
            iv = self.iv_vsa
            line_array = self.line_array_vsa
        else:
            iv = self.iv
            line_array = self.line_array

        i = 0
        for k, q in enumerate( self.qvector ):
            if not is_to_plot( self, k, q, very_small_angle_only=very_small_angle_only ):
                continue

            X = np.ones( len(self.Y) ) * q
            ax1.plot( X, self.Y, line_array[i,:], color='red', alpha=0.2 )
            i += 1

        for k, j in enumerate( self.Y ):
            x = self.qvector[iv]
            y = np.ones( len(iv) ) * j
            z = line_array[:,k]
            ax1.plot( x, y, z, color='yellow', alpha=0.5 )

            if ax2 is not None:
                ax2.plot( x, z, color='yellow' )

                if self.average_curve is not None:
                    if k == len(self.Y)//2:
                        a_curve = self.average_curve * ( self.fg * j + 1 )
                        ax2.plot( x, a_curve, color='green' )
       


# denss/__init__.py

try:
    from ._version import __version__
except ImportError:
    __version__ = ''

from .core import *  # Import EVERYTHING from core.py
from .options import * # Import EVERYTHING from options.py

__all__ = [] # Initialize __all__ as an empty list

# Dynamically populate __all__ with names from core.py and options.py
import inspect
from . import core, options # Import core and options modules to inspect
modules_to_inspect = [core, options] # List of modules to inspect

for module in modules_to_inspect:
 for name, obj in inspect.getmembers(module):
     if not name.startswith('_'): # Exclude private names (starting with _)
         if inspect.isclass(obj) or inspect.isfunction(obj): # Include classes and functions
             __all__.append(name)



"""
    Tools/MapperSingleton.py

    Copyright (c) 2023, SAXS Team, KEK-PF
"""
from molass_legacy.Trimming.PreliminaryRecognition import PreliminaryRecognition
from molass_legacy.Mapping.MapperConstructor import create_mapper

mapper = None
sd = None

def get_mapper(sd_orig):
    global mapper, sd

    if mapper is None:
        pre_recog = PreliminaryRecognition(sd_orig)
        sd = sd_orig._get_analysis_copy_impl(pre_recog)
        mapper = create_mapper(None, sd, sd_orig, pre_recog)

    return mapper, sd



"""
    EmbedCushionUtils.py

    Copyright (c) 2023, SAXS Team, KEK-PF
"""
def get_caller_attr(caller, attr_name, init_value):
    try:
        attr_object = caller.__getattribute__(attr_name)
    except:
        caller.__setattr__(attr_name, init_value)
        attr_object = init_value
    return attr_object



# coding: utf-8
"""
    Rgg.RggUtils.py

    Copyright (c) 2021, SAXS Team, KEK-PF
"""
import numpy as np

def normal_pdf(x, mu, sigma):
    return np.exp(-1/2*((x - mu)/sigma)**2)/(sigma*np.sqrt(np.pi))

def convert_to_probabilitic_data(x, y, rg, max_y=None, num_precision=100):
    data = []
    if max_y is None:
        max_y = np.max(y)
    for x_, y_, rg_ in zip(x, y, rg):
        data += [(x_, rg_)]*int(y_/max_y*num_precision)
    return np.array(data)

"""
c.f.
https://stackoverflow.com/questions/43926473/convert-plot-to-a-surface-plot-matplotlib
"""
def plot_histogram_2d(ax, x_, y_, rg, max_y, num_precision=100):
    z_ = y_/max_y*num_precision
    y_ = rg
    i = np.arange(len(x_))
    ii, jj = np.meshgrid(i, [0,1])
    xx = x_[ii]
    yy = y_[ii]
    zz = z_[jj]
    zz[1,:] = z_
    ax.plot_surface(xx, yy, zz, color='green', alpha=0.3, edgecolor='green')
    ax.plot(x_, y_, z_, color='orange')



"""
    OurPlotUtils.py

    Copyright (c) 2019-2020, Masatsuyo Takahashi, KEK-PF
"""

def draw_as_image(ax, from_fig, from_ax, exp_elements=False):
    """
    https://stackoverflow.com/questions/4325733/save-a-subplot-in-matplotlib
    https://stackoverflow.com/questions/35286540/display-an-image-with-python
    """
    import io
    import matplotlib.image as mpimg
    buffer = io.BytesIO()
    if exp_elements:
        """
        see also:
        https://matplotlib.org/3.1.0/api/_as_gen/matplotlib.axes.Axes.get_window_extent.html
        https://matplotlib.org/3.1.0/api/_as_gen/matplotlib.axes.Axes.get_tightbbox.html#matplotlib.axes.Axes.get_tightbbox
        """
        renderer = from_fig.canvas.get_renderer()
        extent = from_ax.get_tightbbox(renderer).transformed(from_fig.dpi_scale_trans.inverted())
    else:
        extent = from_ax.get_window_extent().transformed(from_fig.dpi_scale_trans.inverted())
    from_fig.savefig(buffer, bbox_inches=extent)
    buffer.seek(0)
    img=mpimg.imread(buffer)
    ax.imshow(img)



"""
    SecTheory.LognormalStudy.py

    Copyright (c) 2023, SAXS Team, KEK-PF
"""
import numpy as np
from scipy.stats import lognorm 

R2PI = np.sqrt(2*np.pi)

def lognormal_pdf(x, mu, sigma):
    return 1/(x*sigma*R2PI) * np.exp(-(np.log(x) - mu)**2/(2*sigma**2))

def demo():
    import molass_legacy.KekLib.DebugPlot as plt

    x = np.linspace(0.01, 10, 200)

    with plt.Dp():
        fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(12,5))
        ax1.set_title("Numpy")
        ax2.set_title("Scipy")

        for mu, style in [(0, "-"), (1, ":")]:
            for sigma in [1, 0.5, 0.25]:
                ax1.plot(x, lognormal_pdf(x, mu, sigma), style, label=r"$\mu=%g, \sigma=%g$" % (mu, sigma))
                ax2.plot(x, lognorm.pdf(x, sigma, mu, np.exp(mu)), style, label=r"$\mu=%g, \sigma=%g$" % (mu, sigma))
                mode = np.exp(mu - sigma**2)
                ax1.plot(mode, lognormal_pdf(mode, mu, sigma), "o", color="red")
                mx = mode + mu
                ax2.plot(mx, lognorm.pdf(mx, sigma, mu, np.exp(mu)), "o", color="red")
                for p, color in [(0.01, "yellow"), (0.99, "cyan")]:
                    px = lognorm.ppf(p, sigma, mu, np.exp(mu))
                    if px < 10:
                        ax2.plot(px, lognorm.pdf(px, sigma, mu, np.exp(mu)), "o", color=color)

        ax1.legend()
        ax2.legend()

        fig.tight_layout()
        plt.show()

if __name__ == '__main__':
    import os
    import sys
    this_dir = os.path.dirname(os.path.abspath(__file__))
    lib_dir = os.path.dirname(this_dir)
    sys.path.append(lib_dir)

    import seaborn
    seaborn.set()
    import molass_legacy.KekLib

    demo()
